#!/usr/bin/python
# -*- coding: utf-8

'''
Author: Uwe Krien (uwe.krien@rl-institut.de)
Changes by:
Responsibility: Uwe Krien (uwe.krien@rl-institut.de)
'''

import os
import logging
import logging.config
from oemof.tools import helpers


def define_logging(inifile='logging.ini', basicpath=None, subdir='log_files'):
    r"""Initialise the logger using the logging.conf file in the local path.

    Several sentences providing an extended description. Refer to
    variables using back-ticks, e.g. `var`.

    Parameters
    ----------
    inifile : string, optional (default: logging.ini)
        Name of the configuration file to define the logger. If no ini-file
        exist a default ini-file will be downloaded from
        'http://vernetzen.uni-flensburg.de/~git/logging_default.ini' and used.
    basicpath : string, optional (default: '.oemof' in HOME)
        The basicpath for different oemof related informations. By default
        a ".oemof' folder is created in your home directory.
    subdir : string, optional (default: 'log_files')
        The name of the subfolder of the basicpath where the log-files are
        stored.

    Notes
    -----
    By default the INFO level is printed on the screen and the debug level
    in a file, but you can easily configure the ini-file.
    Every module that wants to create logging messages has to import the
    logging module. The oemof logger module has to be imported once to
    initialise it.

    Examples
    --------
    To define the default logge you have to import the python logging library
    and this function. The first logging message should be the path where the
    log file is saved to.

    >>> import logging
    >>> from oemof.tools import logger
    >>> logger.define_logging() # doctest: +SKIP
    17:56:51-INFO-Path for logging: /HOME/.oemof/log_files
    ...
    >>> logging.debug("Hallo")

    """
    url = 'http://vernetzen.uni-flensburg.de/~git/logging_default.ini'
    if basicpath is None:
        basicpath = helpers.get_basic_path()
    logpath = helpers.extend_basic_path(subdir)
    log_filename = os.path.join(basicpath, 'logging.ini')
    if not os.path.isfile(log_filename):
        helpers.download_file(log_filename, url)
    logging.config.fileConfig(os.path.join(basicpath, 'logging.ini'))
    logger = logging.getLogger('simpleExample')
    logger.debug('*********************************************************')
    logging.info('Path for logging: %s' % logpath)
    try:
        check_git_branch()
    except:
        check_version()


def check_version():
    """Returns the actual version number of the used oemof version."""
    # TODO : Please add this feature if you know how.
    pass


def check_git_branch():
    '''
    Passes the used brance and commit to the logger
    '''
    path = os.path.join(os.path.dirname(
        os.path.realpath(__file__)), os.pardir,
        os.pardir, '.git')

    # Reads the name of the branch
    f_branch = os.path.join(path, 'HEAD')
    f = open(f_branch, "r")
    first_line = f.readlines(1)
    name_full = first_line[0].replace("\n", "")
    name_branch = name_full.replace("ref: refs/heads/", "")
    f.close()

    # Reads the code of the last commit used
    f_commit = os.path.join(path, 'refs', 'heads', name_branch)
    f = open(f_commit, "r")
    last_commit = f.read(8)
    f.close()

    logging.info("Used oemof version: {0}@{1}".format(
        last_commit,
        name_branch))

if __name__ == '__main__':
    import doctest

    OC = doctest.OutputChecker
    class AEOutputChecker(OC):
        def check_output(self, want, got, optionflags):
            from re import sub
            if optionflags & doctest.ELLIPSIS:
                want = sub(r'\[\.\.\.\]', '...', want)
            return OC.check_output(self, want, got, optionflags)

    doctest.OutputChecker = AEOutputChecker
    doctest.testmod(verbose=True, optionflags=doctest.ELLIPSIS)


from . import Transport


class Simple(Transport):
    """
    Simple Transport connects two busses with a constant efficiency

    Parameters
    ----------
    eta : float
        Constant efficiency of the transport.
    in_max : float
        Maximum input the transport can handle, in $MW$.
    out_max : float
        Maximum output which can possibly be obtained when using the transport,
        in $MW$.
    """
    optimization_options = {}

    def __init__(self, **kwargs):
        # TODO: Add check for what combination of parameters is given and
        #       calculate the missing ones accordingly. Also write down the
        #       relationshiph between the three parameters in the doctstring.
        super().__init__(**kwargs)
        self.eta = kwargs.get('eta')
        self.in_max = kwargs.get('in_max')
        self.out_max = kwargs.get('out_max')

        if(self.in_max is None and self.out_max is not None):
            self.in_max = self.out_max / self.eta
        if(self.out_max is None and self.in_max is not None):
            self.out_max = self.in_max * self.eta



from .. import Entity


class Bus(Entity):
    """
    The other type of entity in an energy system graph (besides
    Components). A Bus acts as a kind of mediator between producers and
    consumers of a commodity of the same kind. As such it has a type,
    which signifies what kind of commodity goes through the bus.

    Parameters
    ----------
    type : string
        the type of the bus. Can be a meaningful value like e.g. "electricity"
        but may be anything that can be tested for equality and is distinct for
        incompatible Buses.
    price : float
        price per unit of type
    balanced : boolean
        if true a busbalance is created, otherwise the busbalance is ignored
    sum_out_limit : float (default: +inf)
        limit of sum of all outflows over the timehorizon
    # TODO: Find a better name for 'sum_out_limit' (global_outflow_limit?)
    excess : boolean
        if true, an optimization variable is created that takes up the slack of
        outflows to keep the busbalance (sum inflows = sum outflows + excess)
    shortage : boolean
        if true, an optimization variable is created that takes the slack of
        inflows to keep the busbalance (sum inflows + shortage = sum outflows)
    excess_costs : float
        costs per unit of excess that is needed to balance the bus
    shortage_costs : float
        costs per unit of shortage that is needed to balance the bus
    """
    optimization_options = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = kwargs.get("type", None)
        self.price = kwargs.get("price", 0)
        self.balanced = kwargs.get("balanced", True)
        self.sum_out_limit = kwargs.get("sum_out_limit", float("+inf"))
        self.results = {}
        self.excess = kwargs.get('excess', True)
        self.shortage = kwargs.get('shortage', False)
        self.excess_costs = kwargs.get('excess_costs', 0)
        self.shortage_costs = kwargs.get('shortage_costs', 10e10)

        if self.excess == True:
           self.create_excess_slack_component()
        if self.shortage == True:
           self.create_shortage_slack_component()

    def create_excess_slack_component(self):
        """
        """
        excess = ExcessSlack(uid=self.uid+'_excess',
                             inputs=[self],
                             costs=self.excess_costs)
        return excess

    def create_shortage_slack_component(self):
        """
        """
        shortage = ShortageSlack(uid=self.uid+'_shortage',
                                 outputs=[self],
                                 costs=self.shortage_costs)
        return shortage

class Component(Entity):
    """
    Components are one specific type of entity comprising an energy system
    graph, the other being Buses. The important thing is, that connections
    in an energy system graph are only allowed between Buses and
    Components and not between Entities of equal subtypes. This class
    exists only to facilitate this distinction and is empty otherwise.

    Parameters
    ----------
    in_max : list
        maximum input of component (power)
    out_max : list
        maximum output of component (power)
    ub_out : list of pandas.series (or array-like)
        The time-depended maximum output of a component. If ub_out is not set
        out_max is used. Contrary to out_max ub_out has to be array-like. If
        ub_out is set out_max is used as the installed capacity and ub_out as
        the time-depended maximum output. It is in charge of the user that
        these values are not inconsistent. You may use max(ub_out) for out_max.
        (power)
    add_out_limit : float
        limit on additional output "capacity" (power)
    capex : float
        Capital expenditure. To get the capital cost it is multiplied with
        out_max (monetory value per power).
    lifetime : float
        lifetime of component (e.g. years)
    wacc : float
        weigted average cost of capital (dimensionless)
    crf : float
        capital recovery factor: (p*(1+p)^n)/(((1+p)^n)-1)
    opex_fix : float
        fixed operational expenditure (e.g. expenses for staff) (monetory
        value over the full time period).
    opex_var : float
        Variable operational expenditure (e.g. spare parts). You can use it to
        define the fuel costs. Fuel cost can be defined in different ways, so
        be aware not to definge them twice.
    co2_var : float
        variable co2 emissions (e.g. t / MWh)
    co2_cap : float
        co2 emissions due to installed power (e.g. t/ MW)
    """
    optimization_options = {}

    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        self.in_max = kwargs.get('in_max')
        self.out_max = kwargs.get('out_max')
        self.ub_out = kwargs.get('ub_out')
        self.add_out_limit = kwargs.get('add_out_limit')
        self.capex = kwargs.get('capex', 0)
        self.lifetime = kwargs.get('lifetime', 20)
        self.wacc = kwargs.get('wacc', 0.05)
        self.opex_var = kwargs.get('opex_var', 0)
        self.opex_fix = kwargs.get('opex_fix', 0)
        self.co2_var = kwargs.get('co2_var', 0)
        self.co2_cap = kwargs.get('co2_cap', 0)

        self.crf = kwargs.get('crf', None)
        if self.crf is None:
            p = self.wacc
            n = self.lifetime
            self.crf = (p*(1+p)**n)/(((1+p)**n)-1)
        self.results = kwargs.get('results', {'in': {},
                                              'out': {}})

class ExcessSlack(Component):
    """A ExcessSlack is a special sink which takes the output slack i.e. excess
    of the bus to that it is connected."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.costs = kwargs.get("costs", 0)

class ShortageSlack(Component):
    """A ShorageSlack is a special source which takes the input slack i.e.
    shortage of the bus to that it is connected."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.costs = kwargs.get("costs", 10e10)


# -*- coding: utf-8 -*-
"""

@author: Simon Hilpert
"""

#Tools to override slow Pyomo problem building



from pyomo.environ import Constraint
from pyomo.environ import Var

import pyomo


def l_constraint(model, name, constraints, *args):
    r"""A replacement for pyomo's Constraint that quickly builds linear
    constraints.

    Instead of:
    model.constraint_name = Constraint(index1,index2,...,rule=f)

    Call:
    l_constraint(model, constraint_name, constraints, index1, index2,...)

    This is a copy of the code from the python module pypsa. Thanks to
    Tom Brown / Jonas Hoersch for implementing it!

    Parameters
    ----------

    model : pyomo.ConcreteModel() / SimpleBlock() instance
        pyomo model or block with constructed components
    name : string
       Name of constraint to be constructed
    constraints: dict
        Constraints is a dictionary of constraints of the form:

        constraints[i] = [[(coeff1, var1), (coeff2, var2),...], sense, constant_term]

        sense is one of "==","<=",">=".

        constant_term is constant rhs of equation

        I.e. variable coefficients are stored as a list of tuples.

    *args :
       arguments passed to the pyomo.Constraint() class.


    """

    setattr(model, name, Constraint(*args, noruleinit=True))

    v = getattr(model, name)

    for i in v._index:
        c = constraints[i]
        v._data[i] = pyomo.core.base.constraint._GeneralConstraintData(None, v)
        v._data[i]._body = pyomo.core.base.expr_coopr3._SumExpression()
        v._data[i]._body._args = [item[1] for item in c[0]]
        v._data[i]._body._coef = [item[0] for item in c[0]]
        v._data[i]._body._const = 0.
        if c[1] == "==":
            v._data[i]._equality = True
            v._data[i]._lower = pyomo.core.base.numvalue.NumericConstant(c[2])
            v._data[i]._upper = pyomo.core.base.numvalue.NumericConstant(c[2])
        elif c[1] == "<=":
            v._data[i]._equality = False
            v._data[i]._lower = None
            v._data[i]._upper = pyomo.core.base.numvalue.NumericConstant(c[2])
        elif c[1] == ">=":
            v._data[i]._equality = False
            v._data[i]._lower = pyomo.core.base.numvalue.NumericConstant(c[2])
            v._data[i]._upper = None
    #v.construct()


def mutate_variable(var=None, value=None, bounds=(0, None), fix=False,
                    index=None):
    r""" Mutates existing pyomo variable, could be used to extend an emtpy
    Var() object or an existing indexed pyomo variable.
    It's possible to overwrite information for existing indices.

    Parameters
    ----------
    var : pyomo.core.base.var.Var() object
       pyomo variable to be extended
    value : list
       values of variable in list
    bounds : list
       lower and upper bounds of variable as tuples in list
    index : type of var._index elements (most likely tuple)
       list of indices for extension (Warning: if existing indices are used,
       data for for these indices will be overwritten.)
    fix : list
       list with boolean elements, if true, variable value is fixed, i.e.
       converted to parameter in opt-problem

    Example
    -------
    extend_variable(om.w, value=[10], bounds=[(0,10)], fix=[True],
                    index=[('bel','sink',155)])
    """

    bounds = dict(zip(index, bounds))
    value = dict(zip(index, value))
    fix = dict(zip(index, fix))

    for i in index:
        var._data[i] = pyomo.core.base.var._VarData(var)
        var._data[i].value = value[i]
        var._data[i].setlb(bounds[i][0])
        var._data[i].setub(bounds[i][1])

        if fix[i]:
            var._data[i].fix()



#!/usr/bin/python
# -*- coding: utf-8

import logging
import pandas as pd
try:
    import matplotlib.pyplot as plt
except:
    logging.warning('Matplotlib does not work.')


class ResultsDataFrame(pd.DataFrame):
    r"""Creates a multi-indexed pandas dataframe from a solph result object
    and holds methods to create subsets of the data.

    Note
    ----
    This is so far only a rough sketch and serves as a base for discussion.

    Parameters
    ----------

    energy_system : class:`Entity <oemof.core.EnergySystem>`
        energy supply system
    """

    def __init__(self, **kwargs):
        # default values if not arguments are passed
        es = kwargs.get('energy_system')

        rows_list = []
        for k, v in es.results.items():
            if ('Bus' in str(k.__class__)):
                for kk, vv in v.items():
                    row = {}
                    row['bus_uid'] = k.uid
                    row['bus_type'] = k.type
                    row['type'] = ('output' if not (isinstance(kk, str)) else
                                   'other')
                    row['obj_uid'] = 'duals' if (k is kk) else (
                                      kk     if (isinstance(kk, str)) else
                                      kk.uid)
                    row['datetime'] = es.time_idx
                    row['val'] = vv
                    rows_list.append(row)
            else:
                if k in v.keys():
                    # self ref. components (results[component][component])
                    for kk, vv in v.items():
                        if(k is kk):
                            # self ref. comp. (results[component][component])
                            row = {}
                            row['bus_uid'] = k.outputs[0].uid
                            row['bus_type'] = k.outputs[0].type
                            row['type'] = 'other'
                            row['obj_uid'] = k.uid
                            row['datetime'] = es.time_idx
                            row['val'] = vv
                            rows_list.append(row)
                        else:
                            # bus inputs (only self ref. components)
                            row = {}
                            row['bus_uid'] = k.outputs[0].uid
                            row['bus_type'] = k.outputs[0].type
                            row['type'] = 'input'
                            row['obj_uid'] = k.uid
                            row['datetime'] = es.time_idx
                            row['val'] = v.get(k.outputs[0])
                            rows_list.append(row)
                else:
                    for kk, vv in v.items():
                        # bus inputs (results[component][bus])
                        row = {}
                        row['bus_uid'] = kk.uid
                        row['bus_type'] = kk.type
                        row['type'] = 'input'
                        row['obj_uid'] = k.uid
                        row['datetime'] = es.time_idx
                        row['val'] = vv
                        rows_list.append(row)

        # split date and value lists to tuples
        tuples = [
            (item['bus_uid'], item['bus_type'], item['type'], item['obj_uid'],
             date, val)
            for item in rows_list for date, val in zip(item['datetime'],
                                                       item['val'])]

        # create multiindexed dataframe
        index = ['bus_uid', 'bus_type', 'type',
                 'obj_uid', 'datetime']

        columns = index + ['val']

        super().__init__(tuples, columns=columns)
        self.set_index(index, inplace=True)
        self.sort_index(inplace=True)

    def slice_by(self, **kwargs):
        r""" Method for slicing the ResultsDataFrame. A subset is returned.

        Parameters
        ----------
        bus_uid : string
        bus_type : string (e.g. "el" or "gas")
        type : string (input/output/other)
        obj_uid: string
        date_from : string
            Start date selection e.g. "2016-01-01 00:00:00". If not set, the
            whole time range will be plotted.
        date_to : string
            End date selection e.g. "2016-03-01 00:00:00". If not set, the
            whole time range will be plotted.

        """
        kwargs.setdefault('bus_uid', slice(None))
        kwargs.setdefault('bus_type', slice(None))
        kwargs.setdefault('type', slice(None))
        kwargs.setdefault('obj_uid', slice(None))
        kwargs.setdefault(
            'date_from', self.index.get_level_values('datetime')[0])
        kwargs.setdefault(
            'date_to', self.index.get_level_values('datetime')[-1])

        # slicing
        idx = pd.IndexSlice

        subset = self.loc[idx[
            kwargs['bus_uid'],
            kwargs['bus_type'],
            kwargs['type'],
            kwargs['obj_uid'],
            slice(pd.Timestamp(kwargs['date_from']),
                  pd.Timestamp(kwargs['date_to']))], :]

        return subset

    def slice_unstacked(self, unstacklevel='obj_uid', **kwargs):
        r"""Method for slicing the ResultsDataFrame. A unstacked
        subset is returned.

        Parameters
        ----------
        unstacklevel : string (default: 'obj_uid')
            Level to unstack the subset of the DataFrame.
        """
        subset = self.slice_by(**kwargs)
        subset = subset.unstack(level=unstacklevel)
        subset.columns = subset.columns.droplevel()
        return subset


class DataFramePlot(ResultsDataFrame):
    r"""Creates plots based on the subset of a multi-indexed pandas dataframe
    of the :class:`ResultsDataFrame class
    <oemof.outputlib.to_pandas.ResultsDataFrame>`.

    Parameters
    ----------
    subset : pandas.DataFrame
        A subset of the results DataFrame.
    ax : matplotlib axis object
        Axis object of the last plot.

    Attributes
    ----------
    subset : pandas.DataFrame
        A subset of the results DataFrame.
    ax : matplotlib axis object
        Axis object of the last plot.
    """

    def __init__(self, **kwargs):
        super(DataFramePlot, self).__init__(**kwargs)
        self.subset = kwargs.get('subset')
        self.ax = kwargs.get('ax')

    def slice_unstacked(self, unstacklevel='obj_uid', **kwargs):
        r"""Method for slicing the ResultsDataFrame. The subset attribute
        will set to an unstacked subset. The self-attribute is returned to
        allow chaining. This method is an extension of the
        :class:`slice_unstacked <ResultsDataFrame.slice_unstacked>` method
        of the `ResultsDataFrame` class (parent class).

        Parameters
        ----------
        unstacklevel : string (default: 'obj_uid')
            Level to unstack the subset of the DataFrame.
        """
        self.subset = super(
            DataFramePlot, self).slice_unstacked(
                unstacklevel='obj_uid', **kwargs)
        return self

    def color_from_dict(self, colordict):
        r""" Method to convert a dictionary containing the components and its
        colors to a color list that can be directly useed with the color
        parameter of the pandas plotting method.

        Parameters
        ----------
        colordict : dictionary
            A dictionary that has all possible components as keys and its
            colors as items.

        Returns
        -------
        list
            Containing the colors of all components of the subset attribute
        """
        tmplist = list(
            map(colordict.get, list(self.subset.columns)))
        tmplist = ['#ff00f0' if v is None else v for v in tmplist]
        if len(tmplist) == 1:
            colorlist = tmplist[0]
        else:
            colorlist = tmplist
        return colorlist

    def set_datetime_ticks(self, tick_distance=None, number_autoticks=3,
                           date_format='%d-%m-%Y %H:%M'):
        r""" Set configurable ticks for the time axis. One can choose the
        number of ticks or the distance between ticks and the format.

        Parameters
        ----------
        tick_distance : real
            The disctance between to ticks in hours. If not set autoticks are
            set (see number_autoticks).
        number_autoticks : int (default: 3)
            The number of ticks on the time axis, independent of the time
            range. The higher the number of ticks is, the shorter should be the
            date_format string.
        date_format : string (default: '%d-%m-%Y %H:%M')
            The string to define the format of the date and time. See
            https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior
            for more information.
        """
        dates = self.subset.index.get_level_values('datetime').unique()
        if tick_distance is None:
            tick_distance = int(len(dates) / number_autoticks) - 1
        self.ax.set_xticks(range(0, len(dates), tick_distance),
                           minor=False)
        self.ax.set_xticklabels(
            [item.strftime(date_format)
             for item in dates.tolist()[0::tick_distance]],
            rotation=0, minor=False)

    def outside_legend(self, reverse=False, plotshare=0.9, **kwargs):
        r""" Move the legend outside the plot. Bases on the ideas of Joe
        Kington. See
        http://stackoverflow.com/questions/4700614/how-to-put-the-legend-out-of-the-plot
        for more information.

        Parameters
        ----------
        reverse : boolean (default: False)
            Print out the legend in reverse order. This is interesting for
            stack-plots to have the legend in the same order as the stacks.
        plotshare : real (default: 0.9)
            Share of the plot area to create space for the legend (0 to 1).
        loc : string (default: 'center left')
            Location of the plot.
        bbox_to_anchor : tuple (default: (1, 0.5))
            Set the anchor for the legend.
        ncol : integer (default: 1)
            Number of columns of the legend.
        handles : list of handles
            A list of handels if they are already modified by another function
            or method. Normally these handles will be automatically taken from
            the artis object.
        lables : list of labels
            A list of labels if they are already modified by another function
            or method. Normally these handles will be automatically taken from
            the artis object.
        Note
        ----
        All keyword arguments (kwargs) will be directly passed to the
        matplotlib legend class. See
        http://matplotlib.org/api/legend_api.html#matplotlib.legend.Legend
        for more parameters.
        """
        kwargs.setdefault('loc', 'center left')
        kwargs.setdefault('bbox_to_anchor', (1, 0.5))
        kwargs.setdefault('ncol', 1)
        handles = kwargs.pop('handles', self.ax.get_legend_handles_labels()[0])
        labels = kwargs.pop('labels', self.ax.get_legend_handles_labels()[1])

        if reverse:
            handles.reverse()
            labels.reverse()

        box = self.ax.get_position()
        self.ax.set_position([box.x0, box.y0, box.width * plotshare,
                              box.height])

        self.ax.legend(handles, labels, **kwargs)

    def plot(self, **kwargs):
        r""" Passing the subset attribute to the pandas plotting method. All
        parameters will be directly passed to pandas.DataFrame.plot(). See
        http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.plot.html
        for more information.

        Returns
        -------
        self
        """
        self.ax = self.subset.plot(**kwargs)
        return self

    def io_plot(self, bus_uid, cdict, line_kwa={}, bar_kwa={}, **kwargs):
        r""" Plotting a combined bar and line plot to see the fitting of in-
        and outcomming flows of a bus balance.

        Parameters
        ----------
        bus_uid : string
            Uid of the bus to plot the balance.
        colordict : dictionary
            A dictionary that has all possible components as keys and its
            colors as items.
        line_kw : dictionary
            Keyword arguments to be passed to the pandas line plot.
        bar_kw : dictionary
            Keyword arguments to be passed to the pandas bar plot.

        Note
        ----
        Further keyword arguments will be passed to the
        :class:`slice_unstacked method <DataFramePlot.slice_unstacked>`.

        Returns
        -------
        handles, labels
            Manipulated labels to correct the unsual construction of the
            stack line plot. You can use them for further maipulations.
        """
        self.ax = kwargs.get('ax', self.ax)

        if self.ax is None:
            print('None')
            fig = plt.figure()
            self.ax = fig.add_subplot(1, 1, 1)

        # Create a bar plot for all input flows
        self.slice_unstacked(bus_uid=bus_uid, type='input', **kwargs)
        self.subset.plot(kind='bar', linewidth=0, stacked=True, width=1,
                         ax=self.ax, color=self.color_from_dict(cdict),
                         **bar_kwa)

        # Create a line plot for all output flows
        self.slice_unstacked(bus_uid=bus_uid, type='output', **kwargs)

        # The following changes are made to have the bottom line on top layer
        # of all lines. Normally the bottom line is the first line that is
        # plotted and will be on the lowest layer. This is difficult to read.
        new_df = pd.DataFrame(index=self.subset.index)
        n = 0
        tmp = 0
        for col in self.subset.columns:
            if n < 1:
                new_df[col] = self.subset[col]
            else:
                new_df[col] = self.subset[col] + tmp
            tmp = new_df[col]
            n += 1
        new_df.sort_index(axis=1, ascending=False, inplace=True)
        colorlist = self.color_from_dict(cdict)
        if isinstance(colorlist, list):
            colorlist.reverse()
        separator = len(colorlist)
        new_df.plot(kind='line', ax=self.ax, color=colorlist,
                    drawstyle='steps-mid', **line_kwa)

        # Adapt the legend to the new oder
        handles, labels = self.ax.get_legend_handles_labels()
        tmp_lab = [x for x in reversed(labels[0:separator])]
        tmp_hand = [x for x in reversed(handles[0:separator])]
        handles = tmp_hand + handles[separator:]
        labels = tmp_lab + labels[separator:]
        labels.reverse()
        handles.reverse()

        self.ax.legend(handles, labels)
        return handles, labels



# -*- coding: utf-8 -*-
"""
Created on Tue Jul 28 12:04:21 2015

@author: uwe
"""
import numpy as np
import logging
import pandas as pd
import os
from math import ceil as round_up
from datetime import time as settime


class electric_building():
    ''

    def __init__(self, bdew=None, **kwargs):
        # slp is of type bdew_elec_slp!
        self.elec_demand = (
            bdew.slp[kwargs['selp_type']] /
            bdew.slp[kwargs['selp_type']].sum(0) *
            kwargs['annual_elec_demand'])
        self.type = kwargs['selp_type']
        self.annual_load = kwargs['annual_elec_demand']

    @property
    def load(self):
        return self.elec_demand


class HeatBuilding():
    ''
    def __init__(self, time_df, temp, **kwargs):
        self.datapath = os.path.join(os.path.dirname(__file__), 'data')
        self.year = time_df.index.year[1000]
        time_df['temp'] = (temp - 273)
        self.heat_demand = self.create_slp(time_df, **kwargs)
        self.type = kwargs['shlp_type']
        self.annual_load = kwargs['annual_heat_demand']

    def temp_geo_series(self, time_df):
        r'''
        A new temperature vector is generated containing a multy-day
        average temperature as needed in the load profile function.

        Notes
        -----
        Equation for the mathematical series of the average tempaerature [1]_:

        .. math::
            T=\frac{T_{t}+0.5\cdot T_{t-1}+0.25\cdot T_{t-2}+
                    0.125\cdot T_{t-3}}{1+0.5+0.25+0.125}

        with :math:`T_t` = Average temperature on the present day
             :math:`T_{t-1}` = Average temperature on the previous day ...

        References
        ----------
        .. [1] `BDEW <https://www.avacon.de/cps/rde/xbcr/avacon/15-06-30_Leitfaden_Abwicklung_SLP_Gas.pdf>`_, BDEW Documentation for heat profiles.
        '''
        tem = time_df['temp'].resample('D', how='mean').reindex(
            time_df.index).fillna(method="ffill")
        return (tem + 0.5 * np.roll(tem, 24) + 0.25 * np.roll(tem, 48) +
                0.125 * np.roll(tem, 72)) / 1.875

    def temp_interval(self, time_df):
        '''
        Appoints the corresponding temperature interval to each temperature in
        the temperature vector.
        '''
        temp = self.temp_geo_series(time_df)
        temp_dict = ({
            -20: 1, -19: 1, -18: 1, -17: 1, -16: 1, -15: 1, -14: 2,
            -13: 2, -12: 2, -11: 2, -10: 2, -9: 3, -8: 3, -7: 3, -6: 3, -5: 3,
            -4: 4, -3: 4, -2: 4, -1: 4, 0: 4, 1: 5, 2: 5, 3: 5, 4: 5, 5: 5,
            6: 6, 7: 6, 8: 6, 9: 6, 10: 6, 11: 7, 12: 7, 13: 7, 14: 7, 15: 7,
            16: 8, 17: 8, 18: 8, 19: 8, 20: 8, 21: 9, 22: 9, 23: 9, 24: 9,
            25: 9, 26: 10, 27: 10, 28: 10, 29: 10, 30: 10, 31: 10, 32: 10,
            33: 10, 34: 10, 35: 10, 36: 10, 37: 10, 38: 10, 39: 10, 40: 10})
        temp_rounded = [round_up(i) for i in temp]
        temp_int = [temp_dict[i] for i in temp_rounded]
        return np.transpose(np.array(temp_int))

    def get_h_values(self, time_df, **kwargs):
        '''Determine the h-values'''
        file = os.path.join(self.datapath, 'shlp_hour_factors.csv')
        hour_factors = pd.read_csv(file, index_col=0)
        hour_factors = hour_factors.query(
            'building_class=={building_class} and shlp_type=="{shlp_type}"'
            .format(**kwargs))

        # Join the two DataFrames on the columns 'hour' and 'hour_of_the_day'
        # or ['hour' 'weekday'] and ['hour_of_the_day', 'weekday'] if it is
        # not a residential slp.
        residential = kwargs['building_class'] > 0

        left_cols = ['hour_of_day'] + (['weekday'] if not residential else [])
        right_cols = ['hour'] + (['weekday'] if not residential else [])

        SF_mat = pd.DataFrame.merge(
            hour_factors, time_df, left_on=left_cols, right_on=right_cols,
            how='outer', left_index=True).sort().drop(
            left_cols + right_cols, 1)

        # Determine the h values
        h = np.array(SF_mat)[np.array(range(0, 8760))[:], (
            self.temp_interval(time_df) - 1)[:]]
        return np.array(list(map(float, h[:])))

    def get_sigmoid_parameter(self, **kwargs):
        ''' Retrieve the sigmoid parameters from the database'''
        file = os.path.join(self.datapath, 'shlp_sigmoid_factors.csv')
        sigmoid = pd.read_csv(file, index_col=0)
        sigmoid = sigmoid.query(
            'building_class=={building_class} and '.format(**kwargs) +
            'shlp_type=="{shlp_type}" and '.format(**kwargs) +
            'wind_impact=={wind_class}'.format(**kwargs))

        A = float(sigmoid['parameter_a'])
        B = float(sigmoid['parameter_b'])
        C = float(sigmoid['parameter_c'])
        D = float(sigmoid['parameter_d']) if kwargs.get(
            'ww_incl', True) else 0
        return A, B, C, D

    def get_weekday_parameter(self, time_df, **kwargs):
        ''' Retrieve the weekdayparameter from the database'''
        file = os.path.join(self.datapath, 'shlp_weekday_factors.csv')
        F_df = pd.read_csv(file, index_col=0)

        F_df = (F_df.query('shlp_type=="{0}"'.format(
            kwargs['shlp_type'])))
        F_df.drop('shlp_type', axis=1, inplace=True)

        F_df['weekdays'] = F_df.index + 1

        return np.array(list(map(float, pd.DataFrame.merge(
            F_df, time_df, left_on='weekdays', right_on='weekday', how='outer',
            left_index=True).sort()['wochentagsfaktor'])))

    def create_slp(self, time_df, **kwargs):
        '''Calculation of the hourly heat demand using the bdew-equations'''
        time_df['weekday'].mask(time_df['weekday'] == 0, 7, True)
        SF = self.get_h_values(time_df, **kwargs)
        [A, B, C, D] = self.get_sigmoid_parameter(**kwargs)
        F = self.get_weekday_parameter(time_df, **kwargs)

        h = (A / (1 + (B / (time_df['temp'] - 40)) ** C) + D)
        KW = (kwargs['annual_heat_demand'] /
              (sum(h * F) / 24))
        return (KW * h * F * SF)

    @property
    def load(self):
        return self.heat_demand


class bdew_elec_slp():
    'Generate electrical standardized load profiles based on the BDEW method.'

    def __init__(self, conn, time_df, periods=None):
        if periods is None:
            self.periods = {
                'summer1': [5, 15, 9, 14],  # summer: 15.05. to 14.09
                'transition1': [3, 21, 5, 14],  # transition1 :21.03. to 14.05
                'transition2': [9, 15, 10, 31],  # transition2 :15.09. to 31.10
                'winter1': [1, 1, 3, 20],  # winter1:  01.01. to 20.03
                'winter2': [11, 1, 12, 31],  # winter2: 01.11. to 31.12
                }
        else:
            self.periods = periods
        self._year = time_df.index.year[1000]
        self.slp_frame = self.all_load_profiles(conn, time_df)

    def all_load_profiles(self, conn, time_df):
        slp_types = ['h0', 'g0', 'g1', 'g2', 'g3', 'g4', 'g5', 'g6', 'l0',
                     'l1', 'l2']
        new_df = self.create_bdew_load_profiles(time_df, slp_types)

        # Add the slp for the industrial group
        new_df['i0'] = self.simple_industrial_heat_profile(time_df)

        new_df.drop(['hour', 'weekday'], 1, inplace=True)
        # TODO: Gleichmäßig normalisieren der i0-Lastgang hat höhere
        # Jahressumme als die anderen.
        return new_df

    def create_bdew_load_profiles(self, time_df, slp_types):
        '''
        Calculates the hourly electricity load profile in MWh/h of a region.
        '''

        # Read standard load profile series from csv file

        selp_series = pd.read_csv('../oemof/demandlib/selp_series.csv')
        tmp_df = selp_series

        index = pd.date_range(
                pd.datetime(2007, 1, 1, 0), periods=2016, freq='15Min')
                # columns=['period', 'weekday'] + slp_types)

        tmp_df.set_index(index, inplace=True)

        # All holidays(0) are set to sunday(7)
        time_df.weekday = time_df.weekday.replace(0, 7)
        new_df = time_df.copy()

        # Create an empty column for all slp types and calculate the hourly
        # mean.
        how = {'period': 'last', 'weekday': 'last'}
        for slp_type in slp_types:
            tmp_df[slp_type] = tmp_df[slp_type].astype(float)
            new_df[slp_type] = 0
            how[slp_type] = 'mean'
        tmp_df = tmp_df.resample('H', how=how)

        # Inner join the slps on the time_df to the slp's for a whole year
        tmp_df['hour_of_day'] = tmp_df.index.hour + 1
        left_cols = ['hour_of_day', 'weekday']
        right_cols = ['hour', 'weekday']
        tmp_df = tmp_df.reset_index()
        tmp_df.pop('index')

        for p in self.periods.keys():
            a = pd.datetime(self._year, self.periods[p][0],
                            self.periods[p][1], 0, 0)
            b = pd.datetime(self._year, self.periods[p][2],
                            self.periods[p][3], 23, 59)
            new_df.update(pd.DataFrame.merge(
                tmp_df[tmp_df['period'] == p[:-1]], time_df[a:b],
                left_on=left_cols, right_on=right_cols,
                how='inner', left_index=True).sort_index().drop(['hour_of_day'], 1))

        return new_df

    def simple_industrial_heat_profile(self, df):
        ''

        # TODO: Remove the hard coded values
        am = settime(7, 0, 0)
        pm = settime(23, 30, 0)

        df['ind'] = 0

        # Day(am to pm), night (pm to am), week day (week),
        # weekend day (weekend)
        week = [1, 2, 3, 4, 5]
        weekend = [0, 6, 7]

        df['ind'].mask(df['weekday'].between_time(am, pm).isin(week), 0.8,
                       True)
        df['ind'].mask(df['weekday'].between_time(pm, am).isin(week), 0.6,
                       True)
        df['ind'].mask(df['weekday'].between_time(am, pm).isin(weekend), 0.9,
                       True)
        df['ind'].mask(df['weekday'].between_time(pm, am).isin(weekend), 0.7,
                       True)

        if df['ind'].isnull().any(axis=0):
            logging.error('NAN value found in industrial load profile')
        return df.pop('ind')

    @property
    def slp(self):
        return self.slp_frame

    @property
    def year(self):
        return self._year



# -*- coding: utf-8 -*-
"""
The module contains different objective expression terms.

@author: Simon Hilpert (simon.hilpert@fh-flensburg.de)
"""
import numpy as np
import logging

def add_opex_var(model, block, ref='output'):
    """ Variable operation expenditure term for linear objective function.

    If reference of opex is `output`:

    .. math:: \\sum_e \\sum_t w_{e, o_{e,1}}(t) \\cdot C_{e, o_{e,1}}(t)

    If reference of opex is `input`:

    .. math:: \\sum_e \\sum_t w_{i_e,e}(t) \\cdot C_{i_e, e}(t)

    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all objects corresponding to one oemof base class
    ref : string
       Reference side on which opex are based on
        (e.g. powerplant MWhth -> input or MWhel -> output)

    Returns
    -------
    Expression
    """
    if block.uids is None:
        block.uids = [obj.uid for obj in block.objs]

    opex_var = {obj.uid: obj.opex_var for obj in block.objs}
    # outputs for cost objs
    if ref == 'output':
        expr = sum(model.w[e, model.O[e][0], t] * opex_var[e]
                   for e in block.uids
                   for t in model.timesteps)

    elif ref == 'input':
        expr = sum(model.w[model.I[e][0], e, t] * opex_var[e]
                   for e in block.uids
                   for t in model.timesteps)
    return(expr)

def add_input_costs(model, block):
    """ Adds costs for usage of input (fuel, elec, etc. ) if not included in
    opex

    .. math:: \\sum_e \\sum_t w_{i_e, e}(t) \\cdot C_{i_e, e}(t)

    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all objects corresponding to one oemof base class

    Returns
    -------
    Expression
    """
    if block.uids is None:
        block.uids = [obj.uid for obj in block.objs]

    input_costs = {}
    for e in block.objs:
        if e.__dict__.get('input_costs', None) is not None:
            input_costs[e.uid] = e.input_costs
        else:
            input_costs[e.uid] = e.inputs[0].price
    # outputs for cost objs
    expr = sum(model.w[model.I[e][0], e, t] * input_costs[e]
               for e in block.uids for t in model.timesteps)

    return(expr)

def add_opex_fix(model, block, ref=None):
    """ Fixed operation expenditure term for linear objective function.

    Parameters
    ----------
    model : OptimizationModel() instance
    block : SimpleBlock()
         block to group all objects corresponding to one oemof base class
    ref : string
        string to check if capex is referred to capacity (storage) or output
        (e.g. powerplant)

    Returns
    -------
    Expression

    """
    if not block.objs:
        expr=0
        print('No objects defined for adding fixed opex to objective.' +
              'No action taken.')
        return(expr)
    else:
        if block.uids is None:
            block.uids = [obj.uid for obj in block.objs]
        uids_inv = set([obj.uid for obj in block.objs
                       if block.optimization_options.get('investment', False)])
        uids = set(block.uids) - uids_inv

        opex_fix = {obj.uid: obj.opex_fix for obj in block.objs}
        if ref == 'output':
            # installed electrical/thermal output: {'pp_chp': 30000,...}
            out_max = {obj.uid: obj.out_max for obj in block.objs}
            expr = 0
            expr += sum(out_max[e][0] * opex_fix[e] for e in uids)
            expr += sum((out_max[e][0] + block.add_out[e]) * opex_fix[e]
                        for e in uids_inv)
        elif ref == 'capacity':
            cap_max = {obj.uid: obj.cap_max for obj in block.objs}
            expr = 0
            expr += sum(cap_max[e] * opex_fix[e] for e in uids)
            expr += sum((cap_max[e] + block.add_cap[e]) * opex_fix[e]
                           for e in uids_inv)
        else:
            print('No reference defined. Please specificy in `add_opex()`!')
        return(expr)


def add_revenues(model, block, ref='output', idx=0):
    """ Revenue term for linear objective function.

    .. math:: \\sum_e \\sum_t w_{e, o_{e,1}}(t) \\cdot R_{e, o_{e,1}}(t)

    Parameters
    ----------
    model : OptimizationModel() instance
    block : SimpleBlock()
         block to group all objects corresponding to one oemof base class
    ref : string
       Reference side for revnues ('output' or 'input' Note: 'input' not
       defined)
    idx : integer
       Integer indicating which output from list to select
       if entity has multiple outputs

    Returns
    -------
    Expression
    """

    if block.uids is None:
        block.uids = [obj.uid for obj in block.objs]

    expr = 0
    if ref == 'output':
        #  if price is already a vector (array) this vector is taken
        output_price = {}
        for e in block.objs:
            if e.__dict__.get('output_price', None)[idx] is not None:
                if isinstance(e.output_price[idx], (float, int, np.integer,
                                               np.float)):
                    output_price[e.uid] = [e.output_price[idx]] * len(model.timesteps)
                else:
                    output_price[e.uid] = e.output_price[idx]
            elif isinstance(e.outputs[idx].price, (float, int, np.integer)):
                output_price[e.uid] = \
                    [e.outputs[idx].price] * len(model.timesteps)
            else:
                output_price[e.uid] = e.outputs[idx].price

        # create expression term
        expr += -sum(model.w[e, model.O[e][idx], t] * output_price[e][t]
                     for e in block.uids for t in model.timesteps)
    else:
        raise NotImplementedError("Referece side 'input' not implemented.")

    return(expr)

def add_curtailment_costs(model, block=None):
    """ Cost term for dispatchable sources in linear objective.

    .. math:: \\sum_e \\sum_ t w^{cut}_e(t) \\cdot C^{cut}_e


    Parameters
    ----------
    model : OptimizationModel() instance
    block : SimpleBlock()
         block to group all objects corresponding to one oemof base class

    Returns
    -------
    Expression
    """

    if not block.objs:
        expr = 0
        logging.warning('No objects for curtailment costs term in objective' +
        ' defined. No action taken.')
        return(expr)

    else:
        if block.uids is None:
            block.uids = [obj.uid for obj in block.objs]
        # get dispatch expenditure for renewable energies with dispatch
        c_curtail = {obj.uid: obj.curtail_costs for obj in block.objs}
        expr = sum(block.curtailment_var[e, t] * c_curtail[e]
                   for e in block.uids for t in model.timesteps)
    return(expr)

def add_capex(model, block, ref='output'):
    """ Add capital expenditure to linear objective.

    If reference is `output` (e.g. powerplants):


    If reference is `capacity` (e.g. storages):


    Parameters
    ----------
    model : OptimizationModel() instance
    block : SimpleBlock()
         block to group all objects corresponding to one oemof base class
    ref : string
        string to check if capex is referred to capacity (storage) or output
        (e.g. powerplant)

    Returns
    -------
    Expression
    """
    if not block.objs:
        expr = 0
        print('No objects for capex objective term defined. No action taken')
        return(expr)
    else:
        if block.uids is None:
            block.uids = [obj.uid for obj in block.objs]

        c_inv = {obj.uid: obj.capex for obj in block.objs}
        crf = {obj.uid: obj.crf for obj in block.objs}

        if ref == 'output':
            expr = sum(block.add_out[e] * crf[e] * c_inv[e]
                       for e in block.uids)
            return(expr)
        elif ref == 'capacity':
            expr = sum(block.add_cap[e] * crf[e] * c_inv[e]
                       for e in block.uids)
            return(expr)
        else:
            print('No reference defined. Please specificy in `add_capex()`')


def add_startup_costs(model, block):
    """ Adds startup costs for components to objective expression

    .. math:: \\sum_{e} \\sum_{t} z^{start}_e(t) \\cdot C^{start}_e

    Parameters
    ----------
    model : OptimizationModel() instance
    block : SimpleBlock()
         block to group all objects corresponding to one oemof base class

    Returns
    -------
    Expression
    """
    if block.objs is None:
        expr = 0
    else:
        if block.uids is None:
            block.uids = [obj.uid for obj in block.objs]

        c_start = {obj.uid: obj.start_costs for obj in block.objs}
        expr = sum(block.z_start[e, t] * c_start[e]
                   for e in block.uids for t in model.timesteps)
    return(expr)

def add_shutdown_costs(model, block):
    """ Adds shutdown costs for components to objective expression

    .. math:: \\sum_{e} \\sum_t z^{stop}_e(t) \\cdot C^{stop}_e

    Parameters
    ----------
    model : OptimizationModel() instance
    block : SimpleBlock()
         block to group all objects corresponding to one oemof base class

    Returns
    -------
    Expression
    """

    if block.uids is None:
        block.uids = [obj.uid for obj in block.objs]

    c_stop = {obj.uid: obj.stop_costs for obj in block.objs}
    expr = sum(block.z_stop[e, t] * c_stop[e]
               for e in block.uids for t in model.timesteps)
    return(expr)

def add_ramping_costs(model, block, grad_direc='positive'):
    """ Add gradient costs for components to linear objective expression.

    .. math::  \\sum_e \\sum_t g^{pos}_e(t) \\cdot C^{g,neg}_e

    .. math:: \\sum_e \\sum_t g^{neg}_e(t) \\cdot C^{g,pos}_e

    Parameters
    ----------
    model : OptimizationModel() instance
    block : SimpleBlock()
         block to group all objects corresponding to one oemof base class
    grad_direc : string
        direction of gradient for which the costs are added to the objective
        expression

    Returns
    -------
    Expression
    """
    if block.uids is None:
        block.uids = [obj.uid for obj in block. objs]

    c_ramp = {obj.uid: obj.ramp_costs.get(grad_direc, obj.ramp_costs)
              for obj in block.objs}
    if grad_direc == 'positive':
        expr = sum(block.grad_pos_var[e, t] * c_ramp[e]
                   for e in block.uids for t in model.timesteps)
    if grad_direc == 'negative' :
        expr = sum(block.grad_neg_var[e, t] * c_ramp[e]
                   for e in block.uids for t in model.timesteps)
    else:
        pass
    return(expr)

def add_excess_slack_costs(model, block=None):
    """ Artificial cost term for excess slack variables.

    .. math:: \\sum_e \\sum_t EXCESS_e(t) \\cdot C^{excess}_e


    Parameters
    ----------
    model : OptimizationModel() instance
    block : SimpleBlock()    block : SimpleBlock()
         block to group all objects corresponding to one oemof base class

    Returns
    -------
    Expression
    """
    c_excess = {e.uid:e.costs for e in block.objs}
    expr = sum(model.w[model.I[e][0], e,   t] * c_excess[e]
               for e in block.uids for t in model.timesteps)
    return(expr)


def add_shortage_slack_costs(model, block=None):
    """ Artificial cost term for shortage slack variables.

    .. math:: \\sum_e \\sum_t SHORTAGE_e(t) \\cdot C^{shortage}_e

    With :math:`e  \\in E` and :math:`E` beeing the set of unique ids for
    all entities grouped inside the attribute `block.objs`.

    Parameters
    ----------
    model : OptimizationModel() instance
    block : SimpleBlock()
         block to group all objects corresponding to one oemof base class

    Returns
    -------
    Expression
    """
    c_shortage = {e.uid: e.costs for e in block.objs}
    expr = sum(model.w[e, model.O[e][0], t] * c_shortage[e]
               for e in block.uids for t in model.timesteps)
    return(expr)





from . import Source


class FixedSource(Source):
    """
    A fixed source only has one output always. The value of the output is fixed
    for all timesteps in the timehorizon of the optimization problem.
    """
    optimization_options = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class DispatchSource(Source):
    """ Dispatch sources only have one output (like FixedSource) but the
    output can be reduced inside the optimization problem.
    """
    optimization_options = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


# TODO: Implement constraints etc. for Commodity()
class Commodity(Source):
    """ The commodity component can be used to model inputs to resource busses.
    At the moment no constraint etc. are implemented for this component.
    """
    optimization_options = {}

    def __init__(self, **kwargs):

        super().__init__(**kwargs)
        self.sum_out_limit = kwargs.get('sum_out_limit', float('+inf'))
        self.out_max = kwargs.get('out_max', [10e10])


from .. import Component

class Sink(Component):
    """
    A Sink is special Component which only consumes some source commodity.
    Therefore its list of outputs has to be either None or empty
    (i.e. logically False).
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.outputs:
            raise ValueError("Sink must not have outputs.\n" +
                             "Got: {0!r}".format([str(x)
                                                 for x in self.outputs]))


class Source(Component):
    """
    The opposite of a Sink, i.e. a Component which only produces and as a
    consequence has no input.

    Parameters
    ----------
    in_max : list
        maximum input of component (e.g. in MW)
    out_max : list
        maximum output of component (e.g. in MW)
    add_out_limit : float
        limit on additional output "capacity" (e.g. in MW)
    capex : float
        capital expenditure (e.g. in Euro / MW )
    lifetime : float
        lifetime of component (e.g. years)
    wacc : float
        weigted average cost of capital (dimensionless)
    crf : float
        capital recovery factor: (p*(1+p)^n)/(((1+p)^n)-1)
    opex_fix : float
        fixed operational expenditure (e.g. expenses for staff)
    opex_var : float
        variable operational expenditure (e.g. spare parts + fuelcosts)
    co2_fix : float
        fixed co2 emissions (e.g. t / MW)
    co2_var : float
        variable co2 emissions (e.g. t / MWh)
    co2_cap : float
        co2 emissions due to installed power (e.g. t/ MW)
    """
    optimization_options = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.val = kwargs.get('val', None)
        self.curtail_costs = kwargs.get('curtail_costs', 0)

        if self.inputs:
            raise ValueError("Source must not have inputs.\n" +
                             "Got: {0!r}".format([str(x)
                                                 for x in self.inputs]))


class Transformer(Component):
    """
    A Transformer is a specific type of Component which transforms
    (possibly m) inputs into (possibly n) outputs. As such neither its
    list of inputs, nor its list of outputs are allowed to be empty.

    Parameters
    ----------
    out_min : list
        minimal output of transformer (e.g. min power output of powerplants)
    in_min : list
        minimal input of transformer (e.g. min fuel consumption of powerplants)
    grad_pos : float
        positive gradient (absolut value between two sequential timesteps)
    grad_neg : float
        negative gradient (absolut value between two sequential timesteps)
    t_min_off : float
        minimal off time in timesteps (e.g. 5 hours)
    t_min_on : float
        minimal on time in timesteps (e.g  5 hours)
    outages : float or array
        Outages of component.
        either: defined timesteps of timehorizon: e.g. [1,4,200]
        or: 0 <= scalar <= 1 as factor of the total timehorizon e.g. 0.05
    input_costs : float
        costs for usage of input (if not included in opex_var)
    start_costs : float
        cost per start up of transformer (only milp models)
    stop_costs : float
        cost per stop up of transformer (only milp models)
    ramp_costs : float
        costs for ramping
    output_price : list
        price for selling to output bus. prices ordered in the order
        of 'outputs'
    eta_min : list
      efficiency of transformer at minimum load for conversion of input
      to output (order of elements corresponding to order of elements
      out outputs,out_min etc.)
    """
    optimization_options = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.inputs:
            raise ValueError("Transformer must have at least one input.\n" +
                             "Got: {0!r}".format([str(x)
                                                 for x in self.inputs]))
        if not self.outputs:
            raise ValueError("Transformer must have at least one output.\n" +
                             "Got: {0!r}".format([str(x)
                                                 for x in self.outputs]))

        parameters = ['out_min', 'in_min', 'grad_pos', 'grad_neg',
                      't_min_off', 't_min_on', 'outages', 'input_costs',
                      'start_costs', 'stop_costs', 'ramp_costs',
                      'output_price', 'eta_min']

        for k in kwargs:
            if k in parameters:
                setattr(self, k, kwargs[k])


class Transport(Component):
    """
    A Transport is a simple connection transporting a commodity from one
    Bus to a different one. It is different from a Transformer in that it
    may not change the type of commodity being transported. But since the
    transfer can still change things about the commodity other than the
    type (loss, gain, time delay, etc.) this class exists to encapsulate
    such changes.
    """
    optimization_options = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if len(self.inputs) != 1:
            raise ValueError("Transport must have exactly one input.\n" +
                             "Got: {0!r}".format([str(x)
                                                 for x in self.inputs]))

        if len(self.outputs) != 1:
            raise ValueError("Transport must have exactly one output.\n" +
                             "Got: {0!r}".format([str(x)
                                                 for x in self.outputs]))





"""

@contact Simon Hilpert (simon.hilpert@fh-flensburg.de)
"""

from collections import UserDict as UD
from functools import singledispatch

import pyomo.environ as po
import logging

from ..tools import helpers
from . import variables as var
from . import linear_mixed_integer_constraints as milc
from . import linear_constraints as lc
from ..core.network.entities import Bus, Component, ExcessSlack, ShortageSlack
from ..core.network.entities.buses import HeatBus
from ..core.network.entities import components as cp
from ..core.network.entities.components import transformers as transformer
from ..core.network.entities.components.sources import (
    Commodity, DispatchSource, FixedSource)
from ..core.network.entities.components.sinks import Simple as Sink
from ..core.network.entities.components import transports


@singledispatch
def assembler(e, om, block):
    """ Assemblers are the functions adding constraints to an optimization
        model for a set of objects.

    This is the most general form of assembler function, called only if no
    other, more specific assemblers have been found. Since we don't know what
    to do in this case, we can only throw a :class:`TypeError`.

    Parameters
    ----------
    entity: An object. Only used to figure out which assembler function to call
            by dispatching on its `type`. Not used otherwise.
            It's a good idea to set this to `None` if the function is called
            directly via :attr:`assembler.registry`.

    om    : The optimization model. Should be an instance of
            :class:`pyomo.ConcreteModel`.

    block : A pyomo block.

    Returns
    -------
    om    : The optimization model passed in as an argument, with additional
            bus balances.
    """
    raise TypeError(
        "Did not find a way to generate optimization constraints for object:" +
        "\n\n {o}\n\n of type:\n\n {t}".format(o=e, t=type(e)))
    return om


class OptimizationModel(po.ConcreteModel):
    """Create Pyomo model of the energy system.

    Parameters
    ----------
    entities : list with all entity objects
    timesteps : list with all timesteps as integer values
    options : nested dictionary with options to set.

    """

    # TODO Cord: Take "next(iter(self.dict.values()))" where the first value of
    #            dict has to be selected

    def __init__(self, energysystem, loglevel=logging.INFO):
        super().__init__()
        logging.basicConfig(format="%(levelname)s:%(message)s", level=loglevel)
        self.entities = energysystem.entities
        self.energysystem = energysystem
        self.timesteps = energysystem.simulation.timesteps
        self.objective_options = energysystem.simulation.objective_options
        self.relaxed = getattr(energysystem.simulation, "relaxed", False)

        self.T = po.Set(initialize=self.timesteps, ordered=True)
        # calculate all edges ([("coal", "pp_coal"),...])
        self.components = [e for e in self.entities
                           if isinstance(e, Component)]
        self.all_edges = self.edges(self.components)
        var.add_continuous(model=self, edges=self.all_edges)

        # group components by type (cbt: components by type)
        cbt = {}
        for c in self.components:
            cbt[type(c)] = cbt.get(type(c), []) + [c]

        self.I = {c.uid: [i.uid for i in c.inputs[:]] for c in self.components
                  if not isinstance(c, cp.Source)}
        self.O = {c.uid: [o.uid for o in c.outputs[:]] for c in self.components
                  if not isinstance(c, cp.Sink)}

        # Add constraints for all components to the model
        self.build_component_constraints(cbt)

        # group buses by type (bbt: buses by type)
        bbt = {}
        for b in [e for e in self.entities if isinstance(e, Bus)]:
            bbt[type(b)] = bbt.get(type(b), []) + [b]

        # Add constraints for all buses to the model
        self.build_bus_constraints(bbt)

        # create objective function
        if not self.objective_options:
            raise ValueError("No objective options defined!")

        logging.info("Building objective function.")
        self.objective_assembler(objective_options=self.objective_options)

    def build_component_constraints(self, cbt):
        logging.info("Building component constraints.")
        for cls in cbt:
            objs = cbt[cls]
            # "call" methods to add the constraints opt. problem
            if objs:  # Should always be nonempty but who knows...
                uids = [e.uid for e in objs]
                # add pyomo block per cls to OptimizationModel instance
                block = po.Block()
                block.uids = po.Set(initialize=uids)
                block.indexset = po.Set(initialize=block.uids*self.T)
                block.objs = objs
                block.optimization_options = cls.optimization_options
                self.add_component(str(cls), block)
                logging.debug("Creating optimization block for omeof " +
                              "classes: " + block.name)
                assembler.registry[cls](e=None, om=self, block=block)

    def build_bus_constraints(self, bbt):
        logging.info("Building bus constraints.")
        for bls in bbt:
            objs = bbt[bls]
            # "call" methods to add the constraints opt. problem
            if objs:
                # add bus block
                block = po.Block()
                uids = [e.uid for e in objs]
                block.uids = po.Set(initialize=uids)
                block.objs = objs
                self.add_component(str(bls), block)
                assembler.registry[bls](e=None, om=self, block=block)

    def default_assembler(self, block):
        """ Method for setting optimization model objects for blocks

        Parameters
        ----------
        self : OptimizationModel() instance
        block : SimpleBlock()
        """

        if block.optimization_options:
            for k in block.optimization_options:
                block.default_optimization_options.update({
                    k: block.optimization_options[k]})
            block.optimization_options = block.default_optimization_options
        else:
            block.optimization_options = block.default_optimization_options

        if (block.optimization_options.get("investment", False) and
                "milp_constr" in block.optimization_options):
            raise ValueError("Component can not be modeled with milp-constr " +
                             "in investment mode!\n" +
                             "Please change `optimization_options`")

        if "milp_constr" in block.optimization_options:
            # create binary status variables for block components
            var.add_binary(self, block, relaxed=self.relaxed)

        # add additional variables (investment mode)
        if block.optimization_options.get("investment", False):
            add_out_limit = {obj.uid: obj.add_out_limit
                             for obj in block.objs}

            def add_out_bound_rule(block, e):
                return (0, add_out_limit[e])
            block.add_out = po.Var(block.uids, within=po.NonNegativeReals,
                                   bounds=add_out_bound_rule)

        for option in block.optimization_options:
            if option not in ["objective", "investment"]:
                block.optimization_options[option](self, block)

    def objective_assembler(self, objective_options):
        """ calls functions to add predefined objective functions

        """

        revenue_objects = objective_options.get("revenue_objects")
        cost_objects = objective_options.get("cost_objects")

        if objective_options.get("function") is None:
            logging.warning("No objective function selected. If you want " +
                            "to build an objective yourself you can" +
                            " ignore this warning.")
        else:
            objective_options["function"](self,
                                          cost_objects=cost_objects,
                                          revenue_objects=revenue_objects)

    def results(self):
        """ Returns a nested dictionary of the results of this optimization
        model.

        The dictionary is keyed by the :class:`Entities
        <oemof.core.network.Entity>` of the optimization model, that is
        :meth:`om.results()[s][t] <OptimizationModel.results>`
        holds the time series representing values attached to the edge (i.e.
        the flow) from `s` to `t`, where `s` and `t` are instances of
        :class:`Entity <oemof.core.network.Entity>`.

        Time series belonging only to one object, like e.g. shadow prices of
        commodities on a certain :class:`Bus
        <oemof.core.network.entities.Bus>`, dispatch values of a
        :class:`DispatchSource
        <oemof.core.network.entities.components.sources.DispatchSource>` or
        storage values of a
        :class:`Storage
        <oemof.core.network.entities.components.transformers.Storage>` are
        treated as belonging to an edge looping from the object to itself.
        This means they can be accessed via
        :meth:`om.results()[object][object] <OptimizationModel.results>`.

        Object attributes holding optimization results, like e.g. `add_cap` for
        storage objects, can be accessed like this:

          :meth:`om.results()[s].add_cap`

        where `s` is a storage object.

        The value of the objective function is stored under the
        :attr:`om.results().objective` attribute.

        Note that the optimization model has to be solved prior to invoking
        this method.
        """
        # TODO: Maybe make the results dictionary a proper object?
        result = UD()
        result.objective = self.objective()
        for entity in self.entities:
            if (isinstance(entity, cp.Transformer) or
                isinstance(entity, cp.Transport) or
                isinstance(entity, cp.Source) or
                    isinstance(entity, ShortageSlack)):
                if entity.outputs:
                    result[entity] = result.get(entity, UD())
                for o in entity.outputs:
                    result[entity][o] = [self.w[entity.uid, o.uid, t].value
                                         for t in self.timesteps]

                for i in entity.inputs:
                    result[i] = result.get(i, {})
                    result[i][entity] = [self.w[i.uid, entity.uid, t].value
                                         for t in self.timesteps]

            if isinstance(entity, cp.sources.DispatchSource):
                result[entity] = result.get(entity, UD())
                # TODO: Why does this use `entity.outputs[0]`?
                result[entity][entity] = [self.w[entity.uid,
                                                 entity.outputs[0].uid,
                                                 t].value
                                          for t in self.timesteps]

            if (isinstance(entity, cp.Sink) or
                isinstance(entity, ExcessSlack)):
                for i in entity.inputs:
                    result[i] = result.get(i, {})
                    result[i][entity] = [self.w[i.uid, entity.uid, t].value
                                         for t in self.timesteps]

            if isinstance(entity, cp.transformers.Storage):
                result[entity] = result.get(entity, UD())
                result[entity][entity] = [
                    getattr(
                        self, str(transformer.Storage)
                        ).cap[entity.uid, t].value
                    for t in self.timesteps]

            block = getattr(self, str(type(entity)))

            for attribute in ["add_cap", "add_out"]:
                values = getattr(block, attribute, None)
                if values:
                    result[entity] = result.get(entity, UD())
                    setattr(result[entity], attribute, values[entity.uid]())

        if hasattr(self, "dual"):
            for bus in getattr(self, str(Bus)).objs:
                if bus.balanced:
                    result[bus] = result.get(bus, {})
                    result[bus][bus] = [
                        self.dual[getattr(self, str(Bus)).balance[
                            (bus.uid, t)]]
                        for t in self.timesteps]

        return result

    def write_lp_file(self, path=None, filename="problem.lp"):
        if path is None:
            path = helpers.extend_basic_path("lp_files")
        self.write(helpers.get_fullpath(path, filename),
                   io_options={"symbolic_solver_labels": True})
        logging.info("LP-file saved to {0}".format(
            helpers.get_fullpath(path, filename)))

    def solve(self, **kwargs):
        r""" Method that takes care of the communication with the solver
        to solve the optimization model.

        Parameters
        ----------
        self : pyomo.ConcreteModel() object

        \**kwargs : keywords
            Possible keys can be set see below
        solver string:
            solver to be used e.g. "glpk","gurobi","cplex"
        debug : boolean
            If True model is solved in debug mode. lp-file is written.
        duals : boolean
            If True, duals and reduced costs are imported from the solver
            results
        verbose : boolean
            If True informations are printed
        solver_io : string
            pyomo solver interface file format: "lp","python","nl", etc.
        solve_kwargs : dict
            Other arguments for the pyomo.opt.SolverFactory.solve() method
            Example : {"solver_io":"lp"}
        solver_cmdline_options : dict
            Dictionary with command line options for solver
            Examples:
            {"mipgap":"0.01"} results in "--mipgap 0.01"
            {"interior":" "} results in "--interior"


        Returns
        -------
        self : solved pyomo.ConcreteModel() instance

        """
        solver = kwargs.get("solver",  self.energysystem.simulation.solver)
        if solver is None:
            solver = "glpk"
        debug = kwargs.get("debug",  self.energysystem.simulation.debug)
        if debug is None:
            debug = False
        duals = kwargs.get("duals",  self.energysystem.simulation.duals)
        if duals is None:
            duals = False
        verbose = kwargs.get("verbose", self.energysystem.simulation.verbose)
        if verbose is None:
            verbose = False
        solver_io = kwargs.get("solver_io", "lp")
        solve_kwargs = kwargs.get("solve_kwargs", {})
        solver_cmdline_options = kwargs.get("solver_cmdline_options", {})

        from pyomo.opt import SolverFactory
        # Create a "dual" suffix component on the instance
        # so the solver plugin will know which suffixes to collect
        if duals is True:
            logging.debug("Setting suffixes for duals & reduced costs.")
            # dual variables (= shadow prices)
            self.dual = po.Suffix(direction=po.Suffix.IMPORT)
            # reduced costs
            self.rc = po.Suffix(direction=po.Suffix.IMPORT)
        # write lp-file
        if debug is True:
            self.write_lp_file()

        # solve instance
        opt = SolverFactory(solver, solver_io=solver_io)
        # set command line options
        options = opt.options
        for k in solver_cmdline_options:
            options[k] = solver_cmdline_options[k]
        # store results
        logging.info("Handing problem to solver and solving.")
        results = opt.solve(self, **solve_kwargs)

        self.solutions.load_from(results)
        if verbose:
            logging.info("**************************************************")
            logging.info("Optimization problem informations from solph")
            logging.info("**************************************************")
            for k in results:
                logging.info("{0}: {1}".format(k, results[k]))
        else:
            logging.debug("**************************************************")
            logging.debug("Optimization problem informations from solph")
            logging.debug("**************************************************")
            for k in results:
                logging.debug("{0}: {1}".format(k, results[k]))
        return results

    def edges(self, components):
        """Method that creates a list with all edges for the objects in
        components.

        Parameters
        ----------
        self :
        components : list of component objects

        Returns
        -------
        edges : list with tupels that represent the edges
        """

        edges = []
        # create a list of tuples
        # e.g. [("coal", "pp_coal"), ("pp_coal", "b_el"),...]
        for c in components:
            for i in c.inputs:
                ei = (i.uid, c.uid)
                edges.append(ei)
            for o in c.outputs:
                ej = (c.uid, o.uid)
                edges.append(ej)
        return(edges)


@assembler.register(Bus)
def _(e, om, block):
    """ Method creates bus balance for all buses.

    The bus model creates all full balance around the energy buses using
    the :func:`lc.generic_bus_constraint` function.
    Additionally it sets constraints to model limits over the timehorizon
    for resource buses using :func:`lc.generic_limit`

    Parameters
    ----------
    see :func:`assembler`.

    Returns
    -------
    om : The optimization model passed in as an argument, with additional
          bus balances.
    """

    # bus balance constraint for energy bus objects
    lc.add_bus_balance(om, block)

    # set limits for buses
    lc.add_global_output_limit(om, block)
    return om


@assembler.register(HeatBus)
def _(e, om, block):
    """ Method creates bus balance for all buses.

    The bus model creates all full balance around the energy buses using
    the :func:`lc.generic_bus_constraint` function.
    Additionally it sets constraints to model limits over the timehorizon
    for resource buses using :func:`lc.generic_limit`

    Parameters
    ----------
    see :func:`assembler`.

    Returns
    -------
    om : The optimization model passed in as an argument, with additional
          bus balances.
    """

    # slack variables that assures a feasible problem
    # get uids for busses that allow excess
    block.excess_uids = [b.uid for b in block.objs if b.excess is True]
    # get uids for busses that allow shortage
    block.shortage_uids = [b.uid for b in block.objs if b.shortage is True]

    # create variables for "slack" of shortage and excess
    if block.excess_uids:
        om.excess_slack = po.Var(block.excess_uids,
                                 om.timesteps,
                                 within=po.NonNegativeReals)
    if block.shortage_uids:
        om.shortage_slack = po.Var(block.shortage_uids,
                                   om.timesteps,
                                   within=po.NonNegativeReals)

    # bus balance constraint for energy bus objects
    lc.add_bus_balance(om, block)

    # set limits for buses
    lc.add_global_output_limit(om, block)
    return om


@assembler.register(transformer.Simple)
def _(e, om, block):
    """ Method containing the constraints functions for simple
    transformer components.

    Constraints are selected by the `optimization_options` variable of
    :class:`Simple`.

    Parameters
    ----------
    see :func:`assembler`.

    Returns
    -------
    see :func:`assembler`.
    """
    # TODO: This should be dependent on objs classes not fixed if assembler
    # method is used by another assemlber method...

    def linear_constraints(om, block):
        lc.add_simple_io_relation(om, block)
        var.set_bounds(om, block, side="output")

    block.default_optimization_options = {
        "linear_constr": linear_constraints}

    om.default_assembler(block)
    return om


@assembler.register(transformer.PostHeating)
def _(e, om, block):
    """ Method containing the constraints functions for simple
    transformer components.

    Constraints are selected by the `optimization_options` variable of
    :class:`Simple`.

    Parameters
    ----------
    see :func:`assembler`.

    Returns
    -------
    see :func:`assembler`.
    """
    # TODO: This should be dependent on objs classes not fixed if assembler
    # method is used by another assemlber method...

    def linear_constraints(om, block):
        lc.add_two_inputs_io_relation(om, block)
        var.set_bounds(om, block, side="output")
        var.set_bounds(om, block, side="input")
        lc.add_postheat_relation(om, block)

    block.default_optimization_options = {
        "linear_constr": linear_constraints}

    om.default_assembler(block)
    return om


@assembler.register(transformer.CHP)
def _(e, om, block):
    """ Method grouping the constraints for simple chp components.

    The method uses the simple_transformer_assembler() method. The
    optimization_options comes from the transfomer.CHP()

    Parameters
    ----------
    See :func:`assembler`.

    Returns
    -------
    See :func:`assembler`.
    """
    def linear_constraints(om, block):
        lc.add_simple_io_relation(om, block)
        lc.add_simple_chp_relation(om, block)
        var.set_bounds(om, block, side="output")

    block.default_optimization_options = {
        "linear_constr": linear_constraints}

    # simple_transformer assebmler for in-out relation, pmin,.. etc.
    om.default_assembler(block)
    return om


@assembler.register(transformer.SimpleExtractionCHP)
def _(e, om, block):
    """Method grouping the constraints for simple chp components.

    Parameters
    ----------
    See :func:`assembler`.

    Returns
    -------
    See :func:`assembler`.
    """
    def linear_constraints(om, block):
        lc.add_simple_extraction_chp_relation(om, block)
        var.set_bounds(om, block, side="output")
        var.set_bounds(om, block, side="input")

    block.default_optimization_options = {
        "linear_constr": linear_constraints}

    # simple_transformer assebmler for in-out relation, pmin,.. etc.
    om.default_assembler(block)
    return om


@assembler.register(transformer.VariableEfficiencyCHP)
def _(e, om, block):
    """Method grouping the constraints for chp components with variable el.
    efficiency.

    Parameters
    ----------
    See :func:`assembler`.

    Returns
    -------
    See :func:`assembler`.
    """
    def linear_constraints(om, block):
        lc.add_eta_total_chp_relation(om, block)

    def milp_constraints(om, block):
        milc.add_variable_linear_eta_relation(om, block)
        milc.set_bounds(om, block, side="output")

    block.default_optimization_options = {
        "linear_constr": linear_constraints,
        "milp_constr": milp_constraints}

    # simple_transformer assebmler for in-out relation, pmin,.. etc.
    om.default_assembler(block)
    return om


@assembler.register(FixedSource)
def _(e, om, block):
    """Method containing the constraints for
    fixed sources.

    Parameters
    ----------
    See :func:`assembler`.

    Returns
    -------
    See :func:`assembler`.
    """
    def linear_constraints(om, block):
        lc.add_fixed_source(om, block)

    block.default_optimization_options = {
        "linear_constr": linear_constraints}

    # simple_transformer assebmler for in-out relation, pmin,.. etc.
    om.default_assembler(block)
    return om


@assembler.register(DispatchSource)
def _(e, om, block):
    """Method containing the constraints for dispatchable sources.

    Parameters
    ----------
    See :func:`assembler`.

    Returns
    -------
    See :func:`assembler`.
    """
    def linear_constraints(om, block):
        lc.add_dispatch_source(om, block)

    block.default_optimization_options = {
        "linear_constr": linear_constraints}

    if block.optimization_options.get("investment", False):
        raise ValueError("Dispatch source + investment is not possible!")

    # simple_transformer assebmler for in-out relation, pmin,.. etc.
    om.default_assembler(block)
    return om


@assembler.register(Sink)
def _(e, om, block):
    """Method containing the constraints for simple sinks

    Simple sinks are modeled with a fixed output value set for the
    variable of the output.

    Parameters
    ----------
    See :func:`assembler`.

    Returns
    -------
    See :func:`assembler`.
    """
    var.set_fixed_sink_value(om, block)
    return om


@assembler.register(Commodity)
def _(e, om, block):
    """Method containing the constraints for commodity

    Comoodity are modeled with a fixed output value set for the
    variable of the output.

    Parameters
    ----------
    See :func:`assembler`.

    Returns
    -------
    See :func:`assembler`.
    """
    lc.add_global_output_limit(om, block)
    return om


@assembler.register(transformer.Storage)
def _(e, om, block):
    """Simple storage assembler containing the constraints for simple
    storage components.

      Parameters
    ----------
    See :func:`assembler`.

    Returns
    -------
    See :func:`assembler`.
    """
    # add capacity variable
    block.cap = po.Var(block.uids, om.timesteps,
                       within=po.NonNegativeReals)

    def linear_constraints(om, block):
        lc.add_storage_balance(om, block)
        var.set_storage_cap_bounds(om, block)
        if not block.optimization_options.get("investment", False):
            var.set_bounds(om, block, side="output")
            var.set_bounds(om, block, side="input")
        else:
            lc.add_storage_charge_discharge_limits(om, block)

    block.default_optimization_options = {
        "linear_constr": linear_constraints}

    if block.optimization_options.get("investment", False):
        block.add_cap = po.Var(block.uids, within=po.NonNegativeReals)

    om.default_assembler(block)
    return(om)


@assembler.register(transports.Simple)
def _(e, om, block):
    """Simple transport assembler grouping the constraints
    for simple transport components

    The method uses the simple_transformer_assembler() method.

    Parameters
    ----------
    See :func:`assembler`.

    Returns
    -------
    See :func:`assembler`.
    """

    # input output relation for simple transport
    lc.add_simple_io_relation(om, block)
    # bounds
    var.set_bounds(om, block, side="output")
    return(om)

@assembler.register(ExcessSlack)
def _(e, om, block):
    """Excess slack assembler grouping the constraints
    for excess slack components.

    Parameters
    ----------
    See :func:`assembler`.

    Returns
    -------
    See :func:`assembler`.
    """

    # bounds
    # var.set_bounds(om, block, side="output")
    return(om)

@assembler.register(ShortageSlack)
def _(e, om, block):
    """Shortage slack assembler grouping the constraints
    for shortage slack components.

    Parameters
    ----------
    See :func:`assembler`.

    Returns
    -------
    See :func:`assembler`.
    """

    # bounds
    # var.set_bounds(om, block, side="output")
    return(om)



from . import Transformer
import logging
import numpy as np


class Simple(Transformer):
    """
    Simple Transformers always have a simple input output relation with a
    constant efficiency

    Parameters
    ----------
    eta : list
       constant efficiency for conversion of input into output (0 <= eta <= 1)
       e.g. eta = [0.4]
    """
    optimization_options = {}

    def __init__(self, **kwargs):
        """
        """
        super().__init__(**kwargs)
        self.eta = kwargs.get('eta', None)


class PostHeating(Simple):
    r"""
    A postheating transformer can transport heat from one HeatBus to another
    eventhough the temperature levels are different.

    The postheating transformer needs an two buses containing the temperature
    attribute as an input and an output. Another nergy input is used to heat up
    the heat flow if the temperature in the output bus is lower. In the list
    of inputs the additional bus is defined first and then the heat flow.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class CHP(Transformer):
    """
    A CombinedHeatPower Transformer always has a simple input output relation
    with a constant efficiency

    Parameters
    ----------
    eta : list
      constant effciency for converting input into output. First element of
      list is used for conversion of input into first element of
      attribute `outputs`. Second element for second element of attribute
      `outputs`. E.g. eta = [0.3, 0.4]

    """
    optimization_options = {}

    def __init__(self, **kwargs):

        super().__init__(**kwargs)
        self.eta = kwargs.get('eta', [None, None])

class VariableEfficiencyCHP(CHP):
    """
    A CombinedHeatPower Transformer with variable electrical efficiency
    Note: The model uses constraints which require binary variables, hence
    objects of this class will results in mixed-integer-linear-problems.

    Parameters
    ----------
    eta_total : float
        total constant efficiency for the transformer
    eta_el : list
        list containing the minimial (first element) and maximal (second element)
        electrical efficiency (0 <= eta_el <= 1)


    """
    optimization_options = {}

    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        self.eta_total = sum(self.eta)

        # calculate minimal
        self.in_min = [self.out_min[0] / self.eta_min[0]]
        self.in_max = [self.out_max[0] / self.eta[0]]

        A = np.array([[1, self.out_min[0]],
                      [1, self.out_max[0]]])
        b = np.array([self.in_min[0], self.in_max[0]])
        self.coeff = np.linalg.solve(A, b)


class SimpleExtractionCHP(CHP):
    """
    Class for combined heat and power unit with extraction turbine and constant
    power to heat coeffcient in backpressure mode

    Parameters
    ----------
    eta_el_cond : float
        constant el. efficiency for transformer in condensing mode
    beta : float
        power loss index
    sigma : float
        power to heat ratio P/Q in backpressure mode
    """
    optimization_options = {}

    def __init__(self, **kwargs):

        super().__init__(**kwargs)
        self.eta_el_cond = kwargs.get('eta_el_cond', self.eta[0])
        self.beta = kwargs.get('beta')
        self.sigma = kwargs.get('sigma')

        if self.in_max is None:
            try:
                self.in_max = [self.out_max[0]/self.eta_el_cond]
            except:
                raise ValueError('Missing attribute "in_max" for object: \n' +
                                 str(type(self)) + 'Auto calculation failed!')
        if self.eta_el_cond is None:
            raise ValueError('Missing attribute "eta_el_cond" for object: \n' +
                             str(type(self)))
        if self.beta is None:
            raise ValueError('Missing attribute "beta" for object: \n' +
                             str(type(self)))
        if self.sigma is None:
            raise ValueError('Missing attribute "sigma" for object: \n' +
                             str(type(self)))


class Storage(Transformer):
    """
    Parameters
    ----------
    cap_max : float
        absolut maximum state of charge if invest=FALSE,
        absolut maximum state of charge of built capacity if invest=TRUE
    cap_min : float
        absolut minimum state of charge
    cap_initial : float, optional
        The state of charge (soc) at timestep 0.
    add_cap_limit : float
        limit of additional installed capacity (only investment models)
    eta_in : float
        efficiency at charging
    eta_out : float
        efficiency at discharging
    cap_loss : float or list/pandas.Series with length of simulation timesteps
        capacity loss per timestep in p/100
    c_rate_in : float
        c-rate for charging (unit is s^-1)
    c_rate_out : float
        c-rate for discharging (unit is s^-1)
    """
    optimization_options = {}

    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        self.cap_max = kwargs.get('cap_max', 0)
        self.cap_min = kwargs.get('cap_min', None)
        self.add_cap_limit = kwargs.get('add_cap_limit', None)
        self.cap_initial = kwargs.get('cap_initial', None)
        self.eta_in = kwargs.get('eta_in', 1)
        self.eta_out = kwargs.get('eta_out', 1)
        self.cap_loss = kwargs.get('cap_loss', 0)
        self.c_rate_in = kwargs.get('c_rate_in', None)
        self.c_rate_out = kwargs.get('c_rate_out', None)

        if not self.optimization_options.get('investment'):
            if self.cap_max == 0:
                logging.info('Storage cap_max set to default value of 0')
            if self.out_max is None:
                try:
                    self.out_max = [self.c_rate_out * self.cap_max]
                except:
                    raise ValueError('Failed to set out_max automatically.' +
                                     'Did you specify c_rate_out and cap_max?')
            if self.in_max is None:
                try:
                    self.in_max = [self.c_rate_in * self.cap_max]
                except:
                    raise ValueError('Failed to set in_max automatically.' +
                                     'Did you specify c_rate_out and cap_max?')

        if self.optimization_options.get('investment'):
            if self.c_rate_in is None:
                raise ValueError('Missing value for c_rate_in!')
            if self.c_rate_out is None:
                raise ValueError('Missing value for c_rate_out!')



from . import Sink


class Simple(Sink):
    """A simple sink. Use this if you do not know which sink to use."""
    optimization_options = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.val = kwargs.get('val', None)



# -*- coding: utf-8 -*-
"""
Created on Fri Jul 24 19:11:38 2015

@author: caro
"""

import numpy as np
import pandas as pd
from oemof.demandlib import energy_buildings as eb


class electrical_demand():
    '''Calculate the electrical demand for a region with different methods.

    This class calculates the electrical demand for a region. Therefore
    several different methods can be applied.

    Parameters
    ----------
    method : {'scale_profile_csv', scale_profile_db',
                  scale_entsoe', calculate_profile'}, required
        Method to calculate the demand for your region.
        Explanation:

        'scale_profile_csv': read only profile from csv and scale it with
            given or calculated demand

        'scale_profile_db': read only profile from database and scale it with
            given or calculated demand

        'scale_entsoe': read entsoe profile from database and scale it with
            given or calculated demand

        'calculate_profile: Calculate profile from the standard load profiles
            of the three demand sectors (households, service, industry) and
            the corresponding annual electric demand.

    Other Parameters
    ----------------
    Required according to chosen method.

    {'scale_profile_csv'} :
        path : str
            '/path/to/file'
        filename : str
            'filename.csv'

    {'scale_profile_db', 'scale_profile_entsoe'} :
        conn :

    {'scale_profile_csv', 'scale_profile_db', 'scale_profile_entsoe'} :
        annual_elec_demand : int
            Annual demand of your region. Works so far only with a given
            value. Calculating the demand from statistic data for the whole
            region can be an option for further development.

    {'calculate_profile'} :
        ann_el_demand_per_sector : list of dictionaries
            Specification of annual electric demand and the corresponding
            standard load profile type (selp_type) for every sector, e.g.
                ann_el_demand_per_sector = [
                    {'ann_el_demand': int or None,
                     'selp_type: {'h0', 'g0', 'g1', 'g2', 'g3', 'g4', 'g5',
                                  'g6', 'i0'}},
                                  ...]
            if ann_el_demand is None, more parameters to calculate the demand
            are necessary: (works so far only if ann_el_demand for every or
                no sector is specified)

        population : int
            Population of your region.

        ann_el_demand_per_person : list of dictionaries
            Specification of the annual electric demand for one household
            according to the household type (from single to four-person
            households), e.g.
                ann_el_demand_per_person = [
                    {'ann_el_demand': int,
                     'household_type': {'one', 'two', 'three', 'four'}},
                     ...]

        household_structure : list of dictionaries
            Number of people living in every household type. Specification
            for your region, e.g.
                household_structure = [
                    {household_members': int,
                     'household_type': {'one', 'two', 'three', 'four'}},
                     ...]

        comm_ann_el_demand_state : int
            Annual electric demand of the service sector of the next bigger
            region, if not given for your region.

        comm_number_of_employees_state : int
            Number of employees in the service sector of the next bigger
            region.

        comm_number_of_employees_region : int
            Number of employees in the service sector of your region.


    Attributes
    ----------
    annual_demand : int
        Included in **kwargs. Given with initialization or calculated or
        to be calculated within this class according to selected method.

    dataframe : pandas dataframe

    elec_demand : array_like

    profile :

    e_slp :

    Returns
    -------

    Notes
    -----
    .. math::

    References
    ----------
    statistics ...

    B. Schachler: Bewertung des Einsatzes von Kraft-Wärme-Kopplungsanlagen
    hinsichtlich der CO2-Emissionen bei wachsendem Anteil Erneuerbarer
    Energien, Masterarbeit, Technische Universität Berlin, 2014

    Examples
    --------
    These are written in doctest format, and should illustrate how to
    use the function.

    z. B. Summe bilden von elec_demand und abgleichen mit ann_el_demand


    '''
    def __init__(self, method, **kwargs):
        self.annual_demand = kwargs.get('annual_elec_demand')
        if self.annual_demand is None:
            self.annual_demand = self.calculate_annual_demand_region()

        self.dataframe = kwargs.get('dataframe')

        self.decider(method, **kwargs)

    def decider(self, method, **kwargs):
        '''
        '''
        if method == 'scale_profile_csv':
            self.profile = self.read_from_csv(path=
                                              kwargs.get('path'),
                                              filename=
                                              kwargs.get('filename'))
            self.elec_demand = self.scale_profile()

        #TODO: implement
        elif method == 'scale_profile_db':
            conn = kwargs.get('conn')
            self.elec_demand = np.array([111, 222])

        #TODO: implement
        elif method == 'scale_entsoe':
            conn = kwargs.get('conn')
            self.elec_demand = np.array([111, 222])

        #TODO: implement industry + def scale_profile() verwenden
        elif method == 'calculate_profile':
            self.conn = kwargs.get('conn')
            self.e_slp = self.read_selp().slp

#            self.hh_e_slp = self.e_slp['h0']
            self.hh_e_slp = self.e_slp[kwargs.get(
                                       'ann_el_demand_per_sector')[0][
                                       'selp_type']]
            self.comm_e_slp = self.e_slp[kwargs.get(
                                         'ann_el_demand_per_sector')[1][
                                         'selp_type']]
            self.ind_e_slp = self.e_slp[kwargs.get(
                                        'ann_el_demand_per_sector')[2][
                                        'selp_type']]

            if kwargs.get('ann_el_demand_per_sector')[0][
                    'ann_el_demand'] is None:
                self.elec_demand = (
                    self.hh_e_slp / self.hh_e_slp.sum() *
                    self.calculate_annual_demand_households(**kwargs) +
                    self.comm_e_slp / self.comm_e_slp.sum() *
                    self.calculate_annual_demand_commerce(**kwargs)) #+
#                    self.ind_e_slp / self.ind_e_slp.sum()) *
#                    self.calculate_annual_demand_sectors(**kwargs))

            else:

                self.elec_demand = (self.hh_e_slp / self.hh_e_slp.sum() *
                                    kwargs.get(
                                        'ann_el_demand_per_sector')[0][
                                        'ann_el_demand'] +
                                    self.comm_e_slp / self.comm_e_slp.sum() *
                                    kwargs.get(
                                        'ann_el_demand_per_sector')[1][
                                        'ann_el_demand'] +
                                    self.ind_e_slp / self.ind_e_slp.sum() *
                                    kwargs.get(
                                        'ann_el_demand_per_sector')[2][
                                        'ann_el_demand'])

        return self.elec_demand

    def read_from_csv(self, **kwargs):
        '''
        read entire demand timeseries or only profile for further
        processing from csv
        '''
        self.profile = pd.read_csv(kwargs.get('path') +
                                   kwargs.get('filename'),
                                   sep=",")

        self.year = 2010  # TODO: year temporarily

        self.profile = self.profile['deu_' + str(self.year)]

        return self.profile

    def read_from_db(self):
        '''
        read entire demand timeseries or only profile for further
        processing from database
        '''
        return

    def read_entsoe(self):
        return

    def read_selp(self):
        self.e_slp = eb.bdew_elec_slp(self.conn, self.dataframe)
        return self.e_slp

    def scale_profile(self):
        '''
        scale a given profile to a given annual demand, which is the sum
        of the single profile values
        '''
        self.elec_demand = (self.profile /
                            self.profile.sum() *
                            self.annual_demand)
        return self.elec_demand

    #TODO: implement??
    def calculate_annual_demand_region(self):
        '''
        calculate annual demand from statistic data
        '''
        self.annual_demand = 50 + 50
        return self.annual_demand

    def calculate_annual_demand_households(self, **kwargs):
        hh_ann_el_demand_per_person = (kwargs.get('household_structure')[0][
            'household_members'] / kwargs.get('household_members_all') *
            kwargs.get('ann_el_demand_per_person')[0]['ann_el_demand'] +
            kwargs.get('household_structure')[1][
                'household_members'] / kwargs.get('household_members_all') *
            kwargs.get('ann_el_demand_per_person')[1]['ann_el_demand'] +
            kwargs.get('household_structure')[2][
                'household_members'] / kwargs.get('household_members_all') *
            kwargs.get('ann_el_demand_per_person')[2]['ann_el_demand'] +
            kwargs.get('household_structure')[3][
                'household_members'] / kwargs.get('household_members_all') *
            kwargs.get('ann_el_demand_per_person')[3]['ann_el_demand'])

        return  kwargs.get('population') * hh_ann_el_demand_per_person

    def calculate_annual_demand_commerce(self, **kwargs):
        return (kwargs.get('comm_ann_el_demand_state') /
                    kwargs.get('comm_number_of_employees_state') *
                    kwargs.get('comm_number_of_employees_region'))


class heat_demand():
    # not implemented yet
    pass



from .. import Bus


class HeatBus(Bus):
    r"""
    A bus containing a temperature and optionally a return temperature.

    Parameters
    ----------
    temperature : float or array-like
        Temperature of the Bus [Kelvin]
    re_temperature : float or array-like
        Return temperature of the Bus [Kelvin]

    """

    def __init__(self, temperature, re_temperature=None, **kwargs):
        super().__init__(**kwargs)
        self.temperature = temperature
        self.re_temperature = re_temperature



# -*- coding: utf-8 -*-
"""
Created on Mon Jul 20 15:53:14 2015

@author: uwe
"""

import pickle
import logging
import os

from oemof.core.network import Entity
from oemof.core.network.entities.components import transports as transport
from oemof.solph.optimization_model import OptimizationModel as OM


class EnergySystem:
    r"""Defining an energy supply system to use oemof's solver libraries.

    Note
    ----
    The list of regions is not necessary to use the energy system with solph.

    Parameters
    ----------
    entities : list of :class:`Entity <oemof.core.network.Entity>`, optional
        A list containing the already existing :class:`Entities
        <oemof.core.network.Entity>` that should be part of the energy system.
        Stored in the :attr:`entities` attribute.
        Defaults to `[]` if not supplied.
    simulation : core.energy_system.Simulation object
        Simulation object that contains all necessary attributes to start the
        solver library. Defined in the :py:class:`Simulation
        <oemof.core.energy_system.Simulation>` class.
    regions : list of core.energy_system.Region objects
        List of regions defined in the :py:class:`Region
        <oemof.core.energy_system.Simulation>` class.
    time_idx : pandas.index, optional
        Define the time range and increment for the energy system. This is an
        optional parameter but might be import for other functions/methods that
        use the EnergySystem class as an input parameter.

    Attributes
    ----------
    entities : list of :class:`Entity <oemof.core.network.Entity>`
        A list containing the :class:`Entities <oemof.core.network.Entity>`
        that comprise the energy system. If this :class:`EnergySystem` is
        set as the :attr:`registry <oemof.core.network.Entity.registry>`
        attribute, which is done automatically on :class:`EnergySystem`
        construction, newly created :class:`Entities
        <oemof.core.network.Entity>` are automatically added to this list on
        construction.
    simulation : core.energy_system.Simulation object
        Simulation object that contains all necessary attributes to start the
        solver library. Defined in the :py:class:`Simulation
        <oemof.core.energy_system.Simulation>` class.
    regions : list of core.energy_system.Region objects
        List of regions defined in the :py:class:`Region
        <oemof.core.energy_system.Simulation>` class.
    results : dictionary
        A dictionary holding the results produced by the energy system.
        Is `None` while no results are produced.
        Currently only set after a call to :meth:`optimize` after which it
        holds the return value of :meth:`om.results()
        <oemof.solph.optimization_model.OptimizationModel.results>`.
        See the documentation of that method for a detailed description of the
        structure of the results dictionary.
    time_idx : pandas.index, optional
        Define the time range and increment for the energy system. This is an
        optional atribute but might be import for other functions/methods that
        use the EnergySystem class as an input parameter.
    """
    def __init__(self, **kwargs):
        for attribute in ['regions', 'entities', 'simulation']:
            setattr(self, attribute, kwargs.get(attribute, []))

        Entity.registry = self
        self.results = kwargs.get('results')
        self.time_idx = kwargs.get('time_idx')

    # TODO: Condense signature (use Buse)
    def connect(self, bus1, bus2, in_max, out_max, eta, transport_class):
        """Create two transport objects to connect two buses of the same type
        in both directions.

        Parameters
        ----------
        bus1, bus2 : core.network.Bus object
            Two buses to be connected.
        eta : float
            Constant efficiency of the transport.
        in_max : float
            Maximum input the transport can handle, in $MW$.
        out_max : float
            Maximum output which can possibly be obtained when using the
            transport, in $MW$.
        transport_class class
            Transport class to use for the connection
        """
        if not transport_class == transport.Simple:
            logging.error('')
            raise(TypeError(
                "Sorry, `EnergySystem.connect` currently only works with" +
                "a `transport_class` argument of" + str(transport.Simple)))
        for bus_a, bus_b in [(bus1, bus2), (bus2, bus1)]:
            uid = ('transport',) + bus_a.uid + bus_b.uid
            transport_class(uid=uid, outputs=[bus_a], inputs=[bus_b],
                            out_max=[out_max], in_max=[in_max], eta=[eta])

    # TODO: Add concept to make it possible to use another solver library.
    def optimize(self, om=None):
        """Start optimizing the energy system using solph.

        Parameters
        ----------
        om : :class:`OptimizationModel <oemof.solph.optimization_model.OptimizationModel>`, optional
            The optimization model used to optimize the :class:`EnergySystem`.
            If not given, an :class:`OptimizationModel
            <oemof.solph.optimization_model.OptimizationModel>` instance local
            to this method is created using the current :class:`EnergySystem`
            instance as an argument.
            You only need to supply this if you want to observe any side
            effects that solving has on the `om`.

        Returns
        -------
        self : :class:`EnergySystem`
        """
        if om is None:
            om = OM(energysystem=self)

        om.solve(solver=self.simulation.solver, debug=self.simulation.debug,
                 verbose=self.simulation.verbose,
                 duals=self.simulation.duals,
                 solve_kwargs=self.simulation.solve_kwargs)

        self.results = om.results()
        return self

    def dump(self, dpath=None, filename=None, keep_weather=True):
        r""" Dump an EnergySystem instance.
        """
        if dpath is None:
            bpath = os.path.join(os.path.expanduser("~"), '.oemof')
            if not os.path.isdir(bpath):
                os.mkdir(bpath)
            dpath = os.path.join(bpath, 'dumps')
            if not os.path.isdir(dpath):
                os.mkdir(dpath)

        if filename is None:
            filename = 'es_dump.oemof'

        pickle.dump(self.__dict__, open(os.path.join(dpath, filename), 'wb'))

        msg = ('Attributes dumped to: {0}'.format(os.path.join(
            dpath, filename)))
        logging.debug(msg)
        return msg

    def restore(self, dpath=None, filename=None):
        r""" Restore an EnergySystem instance.
        """
        logging.info(
            "Restoring attributes will overwrite existing attributes.")
        if dpath is None:
            dpath = os.path.join(os.path.expanduser("~"), '.oemof', 'dumps')

        if filename is None:
            filename = 'es_dump.oemof'

        self.__dict__ = pickle.load(open(os.path.join(dpath, filename), "rb"))
        msg = ('Attributes restored from: {0}'.format(os.path.join(
            dpath, filename)))
        logging.debug(msg)
        return msg


class Region:
    r"""Defining a region within an energy supply system.

    Note
    ----
    The list of regions is not necessary to use the energy system with solph.

    Parameters
    ----------
    entities : list of core.network objects
        List of all objects of the energy system. All class descriptions can
        be found in the :py:mod:`oemof.core.network` package.
    name : string
        A unique name to identify the region. If possible use typical names for
        regions and english names for countries.
    code : string
        A short unique name to identify the region.
    geom : shapely.geometry object
        The geometry representing the region must be a polygon or a multi
        polygon.

    Attributes
    ----------
    entities : list of core.network objects
        List of all objects of the energy system. All class descriptions can
        be found in the :py:mod:`oemof.core.network` package.
    name : string
        A unique name to identify the region. If possible use typical names for
        regions and english names for countries.
    geom : shapely.geometry object
        The geometry representing the region must be a polygon or a multi
        polygon.
    """
    def __init__(self, **kwargs):
        self.entities = []  # list of entities
        self.add_entities(kwargs.get('entities', []))

        self.name = kwargs.get('name')
        self.geom = kwargs.get('geom')
        self._code = kwargs.get('code')

    # TODO: oder sollte das ein setter sein? Yupp.
    def add_entities(self, entities):
        """Add a list of entities to the existing list of entities.

        For every entity added to a region the region attribute of the entity
        is set

        Parameters
        ----------
        entities : list of core.network objects
            List of all objects of the energy system that belongs to area
            covered by the polygon of the region. All class descriptions can
            be found in the :py:mod:`oemof.core.network` package.
        """

        # TODO: prevent duplicate entries
        self.entities.extend(entities)
        for entity in entities:
            if self not in entity.regions:
                entity.regions.append(self)

    @property
    def code(self):
        """Creating a short code based on the region name if no code is set."""
        if self._code is None:
            name_parts = self.name.replace('_', ' ').split(' ', 1)
            self._code = ''
            for part in name_parts:
                self._code += part[:1].upper() + part[1:3]
        return self._code


class Simulation:
    r"""Defining the simulation related parameters according to the solver lib.

    Parameters
    ----------
    solver : string
        Name of the solver supported by the used solver library.
        (e.g. 'glpk', 'gurobi')
    debug : boolean
        Set the chosen solver to debug (verbose) mode to get more information.
    verbose : boolean
        If True, solver output etc. is streamed in python console
    duals : boolean
        If True, results of dual variables and reduced costs will be saved
    objective_options : dictionary
        'function': function to use from
                    :py:mod:`oemof.solph.predefined_objectives`
        'cost_objects': list of str(`class`) elements. Objects of type  `class`
                        are include in cost terms of objective function.
        'revenue_objects': list of str(`class`) elements. . Objects of type
                           `class` are include in revenue terms of
                           objective function.
    timesteps : list or sequence object
         Timesteps to be simulated or optimized in the used library
    relaxed : boolean
        If True, integer variables will be relaxed
        (only relevant for milp-problems)
    fast_build : boolean
        If True, the standard way of pyomo constraint building is skipped and
        a different function is used.
        (Warning: No guarantee that all expected 'standard' pyomo model
        functionalities work for the constructed model!)
    """
    def __init__(self, **kwargs):
        ''
        self.solver = kwargs.get('solver', 'glpk')
        self.debug = kwargs.get('debug', False)
        self.verbose = kwargs.get('verbose', False)
        self.objective_options = kwargs.get('objective_options', {})
        self.duals = kwargs.get('duals', False)
        self.timesteps = kwargs.get('timesteps')
        self.relaxed = kwargs.get('relaxed', False)
        self.fast_build = kwargs.get('fast_build', False)
        self.solve_kwargs = kwargs.get('solve_kwargs', {})

        if self.timesteps is None:
            raise ValueError('No timesteps defined!')



# -*- coding: utf-8 -*-
"""
The module contains linear mixed integer constraints.

@author: Simon Hilpert (simon.hilpert@fh-flensburg.de)
"""
import pyomo.environ as po


def set_bounds(model, block, side="output"):
    """ Set upper and lower bounds via constraints.

    The bounds are set with constraints using the binary status variable
    of the components. The mathematical formulation is as follows:

    If side is `output`:

    .. math::  w_{e, o_{e,1}}(t) \\leq \\overline{W}_{e, o_{e,1}} \\cdot y_e(t), \
    \\qquad \\forall e, \\forall t

    .. math::  w_{e, o_{e,1}}(t) \\geq \\underline{W}_{e, o_{e,1}} \\cdot y_e(t), \
    \\qquad \\forall e, \\forall t

    If side is `input`:

    .. math::  w_{i_e, e}(t) \\leq \\overline{W}_{i_e, e} \\cdot y_e(t), \
    \\qquad \\forall e, \\forall t

    .. math:: w_{i_e, e}(t) \\geq \\underline{W}_{i_e, e} \\cdot y_e(t), \
    \\qquad \\forall e, \\forall t


    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
        Bounds are altered at model attributes (variables) of `model`
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class
    side : string
       string to select on which side the bounds should be set
       (`ìnput`, `output`)


    """
    if block.objs is None:
        raise ValueError("No objects defined. Please specify objects for \
                         which bounds should be set.")
    if side == "output":
        out_max = {obj.uid: obj.out_max for obj in block.objs}

        # set upper bounds
        def output_ub_rule(block, e, t):
            lhs = model.w[e, model.O[e][0], t]
            rhs = block.y[e, t] * out_max[e][0]
            return(lhs <= rhs)
        block.maximum_output = po.Constraint(block.indexset,
                                             rule=output_ub_rule)

        out_min = {obj.uid: obj.out_min for obj in block.objs}
        # set lower bounds
        def output_lb_rule(block, e, t):
            lhs = block.y[e, t] * out_min[e][0]
            rhs = model.w[e, model.O[e][0], t]
            return(lhs <= rhs)
        block.minimum_output = po.Constraint(block.indexset,
                                             rule=output_lb_rule)

    if side == "input":
        in_max = {obj.uid: obj.in_max for obj in block.objs}

        # set upper bounds
        def input_ub_rule(block, e, t):
            lhs = model.w[model.I[e][0], e, t]
            rhs = block.y[e, t] * in_max[e][0]
            return(lhs <= rhs)
        block.maximum_input = po.Constraint(block.indexset, rule=input_ub_rule)

        in_min = {obj.uid: obj.in_min for obj in block.objs}
        # set lower bounds
        def input_lb_rule(block, e, t):
            lhs = block.y[e, t] * in_min[e][0]
            rhs = model.w[model.I[e][0], e, t]
            return(lhs <= rhs)
        block.minimum_input = po.Constraint(block.indexset, rule=input_lb_rule)

def add_variable_linear_eta_relation(model, block):
    """ Adds constraint for input-output relation for all
    units with variable efficiency grouped in `block.objs`.

    The mathematical formulation for the constraint is as follows:

    .. math:: w_{i_e, e}(t) = y_e(t) \\cdot c_1 + c_2 \\cdot w_{e, o_{e,1}}(t), \
    \\qquad \\forall e, \\forall t


    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class

    """
    if not block.objs or block.objs is None:
        raise ValueError('No objects defined. Please specify objects for \
                          which backpressure chp constraints should be set.')

    c = {obj.uid: obj.coeff for obj in block.objs}

    def variable_linear_eta_rule(block, e, t):
        lhs = model.w[model.I[e][0], e, t]
        rhs = block.y[e,t]*c[e][0] + c[e][1] * model.w[e, model.O[e][0], t]
        return(lhs == rhs)
    block.variable_linear_eta_relation = po.Constraint(block.indexset,
                                            rule=variable_linear_eta_rule,
                                            doc="INFLOW = Y*c1 + c2*OUTFLOW_1")

def add_output_gradient_constraints(model, block, grad_direc="both"):
    """ Creates constraints to model the output gradient for milp-models.

    If gradient direction is `positive`:

    .. math:: w_{e,o_{e,1}}(t) - w_{e,o_{e,1}}(t-1) \\leq \\overline{G}^{pos}_{e_{o,1}} + \
    \\underline{W}_{e, o_{e,1}} \\cdot (1 - y_e(t))

    If gradient direction is `negative`:

    .. math:: w_{e,o_{e,1}}(t-1) - w_{e,o_{e,1}}(t) \\leq \\overline{G}^{neg}_{e_{o,1}} + \
    \\underline{W}_{e, o_{e,1}} \\cdot (1 - y_{e}(t-1))


    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class
    grad_direc : string
         direction of gradient ("both", "positive", "negative")

    References
    ----------
    .. [1] M. Steck (2012): "Entwicklung und Bewertung von Algorithmen zur
       Einsatzplanerstelleung virtueller Kraftwerke", PhD-Thesis,
       TU Munich, p.38
    """
    if block.objs is None:
        raise ValueError("No objects defined. Please specify objects for \
                          which bounds should be set.")

    out_min = {obj.uid: obj.out_min for obj in block.objs}
    grad_pos = {obj.uid: obj.grad_pos for obj in block.objs}

    # TODO: Define correct boundary conditions for t-1 of time
    def grad_pos_rule(block, e, t):
        if t > 1:
            return(model.w[e, model.O[e][0], t] - \
               model.w[e, model.O[e][0], t-1] <=  \
               grad_pos[e] + out_min[e][0] * (1 -block.y[e, t]))
        else:
            return(po.Constraint.Skip)

    grad_neg = {obj.uid: obj.grad_neg for obj in block.objs}
    # TODO: Define correct boundary conditions for t-1 of time horizon
    def grad_neg_rule(block, e, t):
        if t > 1:
            lhs = model.w[e, model.O[e][0], t-1] - model.w[e, model.O[e][0], t]
            rhs =  grad_neg[e] + \
                   out_min[e][0] * (1 -block.y[e, t-1])
            return(lhs <=  rhs)

        else:
            return(po.Constraint.Skip)

    # positive gradient
    if grad_direc == "positive" or grad_direc == "both":
        block.milp_gradient_pos = po.Constraint(block.indexset,
                                                rule=grad_pos_rule)
    # negative gradient
    if grad_direc == "negative" or grad_direc == "both":
        block.milp_gradient_neg = po.Constraint(block.indexset,
                                                rule=grad_neg_rule)


def add_startup_constraints(model, block):
    """ Creates constraints to model the start up of a components.

    The mathematical formulation of constraint is as follows:

    .. math::  y_e(t) - y_e(t-1) \\leq z^{start}_e(t), \\qquad \
        \\forall e, \\forall t

    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
        block to group all constraints and variables etc., block corresponds
        to one oemof base class

    """

    if block.objs is None:
        raise ValueError("No objects defined. Please specify objects for \
                          which constraints should be set.")

    # create binary start-up variables for objects
    block.z_start = po.Var(block.uids, model.timesteps, within=po.Binary)

    def start_up_rule(model, e, t):
        if t >= 1:
            try:
                lhs = block.y[e, t] - block.y[e, t-1] - block.z_start[e, t]
                rhs = 0
                return(lhs <= rhs)
            except:
                raise AttributeError('Constructing startup constraints for' +
                                     ' component with uid: %s went wrong', e)
        else:
            # TODO: Define correct boundary conditions
            return(po.Constraint.Skip)
    block.start_up = po.Constraint(block.indexset, rule=start_up_rule)


def add_shutdown_constraints(model, block):
    """ Creates constraints to model the shut down of a component.

    The mathematical formulation for the constraint is as follows:

    .. math::  y_e(t-1) - y_e(t) \\leq z^{stop}_e(t), \\qquad \
    \\forall e, \\forall t


    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class

    """

    if block.objs is None:
        raise ValueError("No objects defined. Please specify objects for \
                          which constraints should be set.")

    # create binary start-up variables for objects
    block.z_stop = po.Var(block.uids, model.timesteps, within=po.Binary)

    def shutdown_rule(block, e, t):
        if t > 1:
            lhs = block.y[e, t-1] - block.y[e, t] - block.z_stop[e, t]
            rhs = 0
            return(lhs <= rhs)
        else:
            # TODO: Define correct boundary conditions
            return(po.Constraint.Skip)
    block.shut_down = po.Constraint(block.indexset, rule=shutdown_rule)

def add_minimum_downtime(model, block):
    """ Adds minimum downtime constraints for for components grouped inside
    `block.objs`.

    The mathematical formulation for constraints is as follows:

     .. math::  (y_e(t-1)-y_e(t)) \\cdot T^{min,off}_e \\leq T^{min,off}_e - \
     \\sum_{\\gamma=0}^{T^{min,off}_e-1} y_e(t+\\gamma)  \
     \\qquad \\forall e, \\forall t \\in [2, t_{max}-t_{min,off}]

    Extra constraints for last timesteps:

    .. math::  (y_e(t-1)-y_e(t)) \\cdot T^{min,off}_e \\leq T^{min,off}_e - \
     \\sum_{\\gamma=0}^{t_{max}-t} y_e(t+\\gamma)  \
     \\qquad \\forall e, \\forall t \\in [t_{max}-t_{min,off}, t_{max}]


    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class

    """
    if not block.objs or block.objs is None:
        raise ValueError('No objects defined. Please specify objects for' +
                          'which minimum downtime constraints should be set.')

    t_min_off = {obj.uid: obj.t_min_off for obj in block.objs}
    t_max = len(model.timesteps)-1

    def minimum_downtime_rule(block, e, t):
        if t <= 1:
            return po.Constraint.Skip
        elif t >= t_max - t_min_off[e]:
            # Adaption for border sections with range(timesteps_max-t)
            lhs = (block.y[e, t-1] - block.y[e, t]) * t_min_off[e]
            rhs = t_min_off[e] - sum(block.y[e, t + p]
                                     for p in range(t_max - t))
            return(lhs <= rhs)
        else:
            lhs = (block.y[e, t-1] - block.y[e, t]) * t_min_off[e]
            rhs = t_min_off[e] - sum(block.y[e, t + p]
                                     for p in range(t_min_off[e]))
            return(lhs <= rhs)
    block.minimum_downtime = po.Constraint(block.indexset,
                                           rule=minimum_downtime_rule)


def add_minimum_uptime(model, block):
    """ Adds minimum uptime constraints for for components in `block`

    The mathematical formulation for constraints is as follows:

     .. math::  (y_e(t) - y_e(t-1)) \\cdot T^{min,on}_e \\leq  \
     \\sum_{\\gamma=0}^{T^{min,on}_e-1} y_e(t+\\gamma)  \
     \\qquad \\forall e, \\forall t \\in [2, t_{max}-t_{min,on}]

     Extra constraint for the last timesteps:

     .. math::  (y_e(t) - y_e(t-1)) \\cdot T^{min,on}_e \\leq  \
      \\sum_{\\gamma=0}^{t_{max}-t} y_e(t+\\gamma)  \
      \\qquad \\forall e, \\forall t \\in [t_{max}-t_{min,on}, t_{max}]


    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class

    """
    if not block.objs or block.objs is None:
        raise ValueError('No objects defined. Please specify objects for' +
                          'which minimum uptime constraints should be set.')

    t_min_on = {obj.uid: obj.t_min_on for obj in block.objs}
    t_max = len(model.timesteps)-1

    def minimum_uptime_rule(block, e, t):
        if t <= 1:
            return po.Constraint.Skip
        elif t >= t_max - t_min_on[e]:
            # Adaption for border sections with range(timesteps_max-t)
            lhs = (block.y[e, t] - block.y[e, t-1]) * t_min_on[e]
            rhs = sum(block.y[e, t + p] for p in range(t_max - t))
            return(lhs <= rhs)
        else:
            lhs = (block.y[e, t] - block.y[e, t-1]) * t_min_on[e]
            rhs = sum(block.y[e, t + p] for p in range(t_min_on[e]))
            return(lhs <= rhs)
    block.minimum_uptime = po.Constraint(block.indexset,
                                          rule=minimum_uptime_rule)



"""
The linear_contraints module contains the pyomo constraints wrapped in
functions. These functions are used by the '_assembler- methods
of the OptimizationModel()-class.

The module frequently uses the dictionaries I and O for the construction of
constraints. I and O contain all components' uids as dictionary keys and the
relevant input input/output uids as dictionary items.

*Illustrative Example*:

    Consider the following example of a chp-powerplant modeled with 4 entities
    (3 busses, 1 component) and their unique ids being stored in a list
    called `uids`:

    >>> uids = ['bus_el', 'bus_th', 'bus_coal','pp_coal']
    >>> I = {'pp_coal': 'bus_coal'}
    >>> O = {'pp_coal': ['bus_el', 'bus_th']}
    >>> print(I['pp_coal'])
    bus_coal


In mathematical notation I, O can be seen as indexed index sets The
elements of the sets are the uids of all components (index: `e`). The the
inputs/outputs uids are the elements of the accessed set by the
component index `e`. Generally the index `e` is the index for the uids-sets
containing the uids of objects for which the constraints are build.
For all mathematical constraints the following definitions hold:

    Inputs:
    :math:`\mathcal{I}_e = \\text{Input-uids of entity } e \\in \mathcal{E}`

    Outputs:
    :math:`\mathcal{O}_e = \\text{All output-uids of entity } e \\in \mathcal{E}`


Simon Hilpert (simon.hilpert@fh-flensburg.de)
"""

import inspect
import logging
import pyomo.environ as po
from pandas import Series as pdSeries
from . import pyomo_fastbuild as pofast


def add_bus_balance(model, block=None):
    """ Adds constraint for the input-ouput balance of bus objects.

    The mathematical formulation for the balance is as follows:

    .. math:: \\sum_{i \\in \mathcal{I}_e} w_{i, e}(t) = \\sum_{o \\in O_e} \
    w_{e, o}(t), \\qquad \\forall e, \\forall t

    With :math:`e  \\in \mathcal{E}_B, t \in \mathcal{T}`

    Parameters
    ----------
    model : OptimizationModel() instance
    block : SimpleBlock()
        block to group all constraints and variables etc., block corresponds
        to one oemof base class

    """
    if not block.objs or block.objs is None:
        raise ValueError('Failed to create busbalance. No busobjects defined!')

    uids = []
    I = {}
    O = {}
    for b in block.objs:
        if b.balanced == True:
            uids.append(b.uid)
            I[b.uid] = [i.uid for i in b.inputs]
            O[b.uid] = [o.uid for o in b.outputs]

    block.balanced_uids = po.Set(initialize=uids)
    block.balanced_uids.construct()

    block.balanced_indexset = po.Set(initialize=block.balanced_uids*model.T)
    block.balanced_indexset.construct()

    if not model.energysystem.simulation.fast_build:
        # component inputs/outputs are negative/positive in the bus balance
        def bus_balance_rule(block, e, t):
            lhs = 0
            lhs = sum(model.w[i, e, t] for i in I[e])
            rhs = sum(model.w[e, o, t] for o in O[e])
            return(lhs == rhs)
        block.balance = po.Constraint(block.balanced_indexset,
                                      rule=bus_balance_rule)

    if model.energysystem.simulation.fast_build:
        balance_dict = {}
        for t in model.timesteps:
            for e in uids:
                tuples = [(1, model.w[i, e, t]) for i in I[e]] + \
                         [(-1, model.w[e, o, t]) for o in O[e]]
                balance_dict[e, t] = [tuples, "==", 0.]
        pofast.l_constraint(block, 'balance', balance_dict,
                            block.balanced_indexset)


def add_simple_io_relation(model, block, idx=0):
    """ Adds constraints for input-output relation as simple function for
    all objects in `block.objs`.

    The mathematical formulation of the input-output relation of a simple
    transformer is as follows:

    .. math:: w_{i_e, e}(t) \cdot \\eta_{i,o_{e,n}} = w_{e, o_{e,n}}(t), \
    \\qquad \\forall e, \\forall t

    With :math:`e  \\in \mathcal{E}` and :math:`\mathcal{E}` beeing
    the set of unique ids for all entities grouped inside the
    attribute `block.objs`.

    Additionally :math:`\mathcal{E} \subset \{\mathcal{E}_{IO}, \mathcal{E}_{IOO}\}`.

    :math:`n` indicates the n-th output of component :math:`e` (arg: idx)


    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class
    idx : integer
      Index to choose which output to select (from list of Outputs: O[e][idx])

    """
    if not block.objs or block.objs is None:
        raise ValueError("No objects defined. Please specify objects for \
                         which the constraints should be build")

    eta = {obj.uid: obj.eta for obj in block.objs}


    # constraint for simple transformers: input * efficiency = output
    def io_rule(block, e, t):
        lhs = model.w[model.I[e][0], e, t] * eta[e][idx] - \
            model.w[e, model.O[e][idx], t]
        return(lhs == 0)

    if not model.energysystem.simulation.fast_build:
        block.io_relation = po.Constraint(block.indexset, rule=io_rule,
                                          doc="INFLOW * efficiency = OUTFLOW_n")

    if model.energysystem.simulation.fast_build:
        io_relation_dict = {(e, t): [[(eta[e][idx],
                                      model.w[model.I[e][0], e, t]),
                                      (-1, model.w[e, model.O[e][idx], t])],
                                     "==", 0.]
                            for e,t in block.indexset}
        pofast.l_constraint(block, 'io_relation', io_relation_dict,
                            block.indexset)


def add_two_inputs_io_relation(model, block):
    r""" Adds constraint for the input-output relation of a post heating
    transformer with two input flows.

    Then the amount of all input flows multiplied with their efficiency will be
    the output flow. The efficiency of the each will be calculated so that the
    sum of the efficiency of both flows might be greater than one.

    The mathematical formulation for the constraint is as follows:

    .. math::
        w_{e,i_{e,1}}(t)\cdot\eta_{e,i_{e,1}}(t)+w_{e,i_{e,2}}(t)\cdot
        \eta_{e,i_{e,2}}(t)=w_{e,o_{e}},\qquad\forall e,\forall t\qquad\qquad

    With :math:`e\in\mathcal{E}` and :math:`\mathcal{E}` beeing the
    set of unique ids for all entities grouped inside the
    attribute `block.objs`.

    Additionally: :math:`\mathcal{E} \subset \mathcal{E}_{IIO}`.

    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class

    """
    if not block.objs or block.objs is None:
        raise ValueError("No objects defined. Please specify objects for \
                         which the constraints should be build")

    eta = {obj.uid: obj.eta for obj in block.objs}

    # constraint for simple transformers: input * efficiency = output
    def io_rule(block, e, t):
        lhs = (model.w[model.I[e][0], e, t] * eta[e][0] +
               model.w[model.I[e][1], e, t] * eta[e][1] -
               model.w[e, model.O[e][0], t])
        return(lhs == 0)

    if not model.energysystem.simulation.fast_build:
        pass
    block.io_relation = po.Constraint(
            block.indexset, rule=io_rule,
            doc="INFLOW_1 * efficiency_1 +  INFLOW_2 * efficiency_2 = OUTFLOW"
            )

    if model.energysystem.simulation.fast_build:
        name = inspect.stack()[0][3]
        logging.warning(
            'No fastbuild constraints defined for function: {0}.'.format(name))


def add_eta_total_chp_relation(model, block):
    """ Adds constraints for input-(output1,output2) relation as
    simple function for all objects in `block.objs`.

    The mathematical formulation of the input-output relation of a simple
    transformer is as follows:

    .. math:: w_{i_e, e}(t) \cdot \\eta^{total}_e = \
    w_{e, o_{e,1}}(t) + w_{e, o_{e,2}}(t), \\qquad \\forall e, \\forall t

    With :math:`e  \\in \mathcal{E}` and :math:`\mathcal{E}` beeing the
    set of unique ids for all entities grouped inside the
    attribute `block.objs`.

    Additionally: :math:`\mathcal{E} \subset \mathcal{E}_{IOO}`.

    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class

    """
    if not block.objs or block.objs is None:
        raise ValueError("No objects defined. Please specify objects for \
                         which the constraints should be build")

    eta_total = {obj.uid: obj.eta_total for obj in block.objs}
    # constraint for simple transformers: input * efficiency = output
    def ioo_rule(block, e, t):
        lhs = model.w[model.I[e][0], e, t] * eta_total[e]
        rhs = model.w[e, model.O[e][0], t] + model.w[e, model.O[e][1], t]
        return(lhs == rhs)
    block.ioo_relation = po.Constraint(block.indexset, rule=ioo_rule,
                             doc="INFLOW * efficiency = OUTFLOW_1 + OUTFLOW_2")


def add_simple_chp_relation(model, block):
    """ Adds constraint for output-output relation for all simple
    combined heat an power units in `block.objs`.

    The mathematical formulation for the constraint is as follows:

    .. math:: \\frac{w_{e, o_{e,1}}(t)}{\\eta_{e, o_{e,1}}(t)} = \
    \\frac{w_{e, o_{e,2}}(t)}{\\eta_{e, o_{e,2}}(t)}, \
    \\qquad \\forall e, \\forall t

    With :math:`e  \\in \mathcal{E}` and :math:`\mathcal{E}` beeing the
    set of unique ids for all entities grouped inside the
    attribute `block.objs`.

    Additionally: :math:`\mathcal{E} \subset \mathcal{E}_{IOO}`.

    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class

    """
    if not block.objs or block.objs is None:
        raise ValueError('No objects defined. Please specify objects for \
                          which backpressure chp constraints should be set.')
    #TODO:
    #  - add possibility of multiple output busses (e.g. for heat and power)
    # efficiencies for simple chps

    eta = {obj.uid: obj.eta for obj in block.objs}

    # additional constraint for power to heat ratio of simple chp comp:
    # P/eta_el = Q/eta_th
    def simple_chp_rule(block, e, t):
        lhs = model.w[e, model.O[e][0], t] / eta[e][0]
        lhs += -model.w[e, model.O[e][1], t] / eta[e][1]
        return(lhs == 0)

    if not model.energysystem.simulation.fast_build:
        block.pth_relation = po.Constraint(block.indexset, rule=simple_chp_rule,
                                           doc="P/eta_el - Q/eta_th = 0")

    if model.energysystem.simulation.fast_build:
        pth_relation_dict = {(e, t): [
                                [(1/eta[e][0], model.w[e, model.O[e][0], t]),
                                 (-1/eta[e][1], model.w[e, model.O[e][1], t])],
                                "==", 0.]
                             for e,t in block.indexset}
        pofast.l_constraint(block, 'pth_relation', pth_relation_dict,
                            block.indexset)


def add_postheat_relation(model, block):
    r""" Adds constraint for the input relation of a post heating device.

    The additional device of the postheat transformer will heat up the flow and
    therefore will deliver an amount of energy. This amount depends on the
    temperature difference between the buses and the temperature difference
    between the input bus and the return flow temperature.

    The mathematical formulation for the constraint is as follows:

    .. math::
        w_{e,i_{e,1}}(t)=w_{e,i_{e,2}}(t)\cdot f,\quad\forall e,\forall t
        \qquad\\
        \\
        f=\frac{T^{flow}_{bus,o}-T^{flow}_{bus,i}}
        {T^{flow}_{bus,i}-T^{return}_{bus,o}}\qquad\qquad\qquad

    With :math:`e\in\mathcal{E}` and :math:`\mathcal{E}` beeing the
    set of unique ids for all entities grouped inside the
    attribute `block.objs`.

    Additionally: :math:`\mathcal{E} \subset \mathcal{E}_{IIO}`.

    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class

    """
    if not block.objs or block.objs is None:
        raise ValueError('No objects defined. Please specify objects for \
                          which post heating constraints should be set.')

    T_coeff = {}
    for e in block.objs:
        out = [o for o in e.outputs[:]]
        T_out = out[0].temperature
        T_re_out = out[0].re_temperature
        T_in = [o for o in e.inputs[:]][1].temperature
        dT_throw = T_out - T_in
        dT_throw[dT_throw < 0] = 0
        dT_in = T_in - T_re_out
        T_coeff[e.uid] = (dT_throw) / (dT_in)

    def postheat_rule(block, e, t):
        lhs = (model.w[model.I[e][1], e, t] * T_coeff[e][t] -
               model.w[model.I[e][0], e, t])
        return(lhs == 0)

    if not model.energysystem.simulation.fast_build:
        block.postheat = po.Constraint(block.indexset, rule=postheat_rule,
                                       doc="Q * f = P_addition")

    if model.energysystem.simulation.fast_build:
        name = inspect.stack()[0][3]
        logging.warning(
            'No fastbuild constraints defined for function: {0}.'.format(name))


def add_simple_extraction_chp_relation(model, block):
    """ Adds constraints for power to heat relation and equivalent output
    for a simple extraction combined heat an power units. The constraints
    represent the PQ-region of the extraction unit and are set for all
    objects in `block.objs`

    The mathematical formulation is as follows:

    For Power/Heat ratio:

    .. math:: w_{e,o_{e,1}}(t) = w_{e, o_{e,2}}(t) \\cdot \\sigma_{e}, \
    \\qquad \\forall e, \\forall t

    .. math:: \\sigma_{e} = \\text{Power to heat ratio of entity } e

    For euivalent power:

    .. math:: w_{i_e,e}(t) = \\frac{w_{e,o_{e,1}}(t) + \\beta_e \\cdot \
    w_{e, o_{e,2}}(t)}{\\eta_{i_e, o_{e,1}}}

    .. math:: \\beta_e = \\text{Power loss index of entity } e

    With :math:`e  \\in \mathcal{E}` and :math:`\mathcal{E}` beeing the
    set of unique ids for all entities grouped inside the
    attribute `block.objs`.

    Additionally: :math:`\mathcal{E} \subset \mathcal{E}_{IOO}`.



    Parameters
    ----------
    model : OptimizationModel() instance
           An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class

    """
    if not block.objs or block.objs is None:
        raise ValueError('No objects defined. Please specify objects for' +
                          'which extraction chp constraints should be set.')

    out_max = {}
    beta = {}
    sigma = {}
    eta_el_cond = {}
    for e in block.objs:
        out_max[e.uid] = e.out_max
        beta[e.uid] = e.beta
        sigma[e.uid] = e.sigma
        eta_el_cond[e.uid] = e.eta_el_cond

    def equivalent_output_rule(block, e, t):
        lhs = model.w[model.I[e][0], e, t]
        rhs = (model.w[e, model.O[e][0], t] +
              beta[e] * model.w[e, model.O[e][1], t]) / eta_el_cond[e]
        return(lhs == rhs)
    block.equivalent_output = po.Constraint(block.indexset,
                                            rule=equivalent_output_rule,
                                            doc='H = (P + Q*beta)/eta_el_cond')
    def power_heat_rule(block, e, t):
        lhs = model.w[e, model.O[e][0], t]
        rhs = sigma[e] *  model.w[e, model.O[e][1], t]
        return(lhs >= rhs)
    block.pth_relation = po.Constraint(block.indexset, rule=power_heat_rule,
                                       doc="P <= sigma * Q")

def add_global_output_limit(model, block=None):
    """ Adds constraints to set limit for variables as sum over the total
    timehorizon for all objects in `block.objs`

    The mathematical formulation is as follows:

    .. math:: \sum_{t \\in \mathcal{T}} \sum_{o \\in \mathcal{O}_e} w_{e, o}(t) \
    \\leq \overline{O}^{global}_e, \\qquad \\forall e \\in E

    With :math:`e  \\in \mathcal{E}` and :math:`\mathcal{E}` beeing the
    set of unique ids for all entities grouped inside the
    attribute `block.objs`.

    Additionally: :math:`\mathcal{E} \subset \{\mathcal{E}_{B},\mathcal{E}_{O}\}`.


    Parameters
    ----------
    model : OptimizationModel() instance
       An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class

    """
    if not block.objs or block.objs is None:
        raise ValueError('Failed to create outputlimit. ' +
                         'No objects defined!')

    limit = {obj.uid: obj.sum_out_limit for obj in block.objs}

    # outputs: {'rcoal': ['coal'], 'rgas': ['gas'],...}
    O = {obj.uid: [o.uid for o in obj.outputs[:]] for obj in block.objs}

    # set upper bounds: sum(yearly commodity output) <= yearly_limit
    def output_limit_rule(block, e):
        lhs = sum(model.w[e, o, t] for t in model.timesteps for o in O[e]) -\
              limit[e]
        # if bus is defined but has not outputs Constraint is skipped
        # TODO: should be logged as well?
        if isinstance(lhs, (int, float)) or limit[e] == float('inf'):
            return(po.Constraint.Skip)
        else:
            return(lhs <= 0)
    block.global_limit = po.Constraint(block.uids, rule=output_limit_rule,
                                       doc="Sum of output <= global_limit")

def add_fixed_source(model, block):
    """ Sets fixed source bounds and constraints for all objects in
    `block.objs`

    The mathematical formulation is as follows:

     .. math::  w_{e,o_e}(t) = V^{norm}_e(t) \\cdot \overline{W}_{e, o_e}, \
     \\qquad \\forall e, \\forall t

    For `investment` for component:

    .. math::  w_{e, o}(t) \\leq (\overline{W}_{e, o_e} + \overline{w}^{add}_{e_o}) \
    \cdot V^{norm}_e(t), \\qquad \\forall e, \\forall t

    .. math:: \overline{w}^{add}_{e_o}  \\leq \overline{W}^{add}_{e_o}, \
    \\qquad \\forall e

    With :math:`e  \\in \mathcal{E}` and :math:`\mathcal{E}` beeing the
    set of unique ids for all entities grouped inside the
    attribute `block.objs`.

    Additionally: :math:`\mathcal{E} \subset \mathcal{E}_{O}`.

    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
        block to group all constraints and variables etc., block corresponds
        to one oemof base class

    """
    if not block.objs or block.objs is None:
        raise ValueError('No objects defined. Please specify objects for' +
                          'which fixed source constriants should be created.')
    # normed value of renewable source (0 <= value <=1)
    val = {}
    out_max = {}
    for e in block.objs:
         out_max[e.uid] = e.out_max
         val[e.uid] = e.val

    # normed value of renewable source (0 <= value <=1)
    val = {obj.uid: obj.val for obj in block.objs}


    if not block.optimization_options.get('investment', False):
        # maximal ouput of renewable source (in general installed capacity)
        out_max = {obj.uid: obj.out_max for obj in block.objs}
        # edges for renewables ([('wind_on', 'b_el'), ...)
        ee = model.edges(block.objs)
        # fixed values for every timestep
        for (e1, e2) in ee:
            for t in model.timesteps:
                # set value of variable
                model.w[e1, e2, t] = val[e1][t] * out_max[e1][0]
                # fix variable value ("set variable to parameter" )
                model.w[e1, e2, t].fix()
    else:
        # set maximum of additional output
        add_out_limit = {obj.uid: obj.add_out_limit for obj in block.objs}
        # loop over all uids (storages) set the upper bound
        for e in block.uids:
            block.add_out[e].setub(add_out_limit[e])

        def invest_rule(block, e, t):
            lhs = model.w[e, model.O[e][0], t]
            rhs = (out_max[e][0] + block.add_out[e]) * val[e][t]
            return(lhs == rhs)
        block.invest = po.Constraint(block.indexset, rule=invest_rule)

def add_dispatch_source(model, block):
    """ Creates dispatchable source bounds/constraints.

    First the maximum value for the output of the source will be set. Then a
    constraint is defined that determines the dispatch of the source. This
    dispatch can be used in the objective function to add cost for dispatch
    of sources.

    The mathemathical formulation of the constraint is as follows:

    .. math:: w^{cut}_{e,o_e}(t) = V^{norm}_e(t) \\cdot \overline{W}_{e,o_e} - \
    w_{e,o_e}(t),  \\qquad \\forall e, \\forall t

    With :math:`e  \\in \mathcal{E}` and :math:`\mathcal{E}` beeing the
    set of unique ids for all entities grouped in the attribute `block.objs`.

    Additionally: :math:`\mathcal{E} \subset \mathcal{E}_{O}`.


    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class

    """
    if not block.objs or block.objs is None:
        raise ValueError('No objects defined. Please specify objects for' +
                          'which dispatch source constaints should be set.')
    # create dispatch var
    block.curtailment_var = po.Var(block.indexset, within=po.NonNegativeReals)

    # normed value of renewable source (0 <= value <=1)
    val = {}
    out_max = {}
    for e in block.objs:
         out_max[e.uid] = e.out_max
         val[e.uid] = e.val

    ee = model.edges(block.objs)
    # fixed values for every timestep
    for (e1, e2) in ee:
        for t in model.timesteps:
            # set upper bound of variable
            model.w[e1, e2, t].setub(val[e1][t] * out_max[e1][0])
    def curtailment_source_rule(block, e, t):
        lhs = block.curtailment_var[e, t]
        rhs = val[e][t] * out_max[e][0] - \
           model.w[e, model.O[e][0], t]
        return(lhs == rhs)
    block.curtailment = po.Constraint(block.indexset,
                                      rule=curtailment_source_rule)


def add_storage_balance(model, block):
    """ Constraint to build the storage balance in every timestep

     The mathematical formulation of the constraint is as follows:

    .. math:: l_e(t) = l_e(t-1) \\cdot (1 - C^{loss}(e)) \
    - \\frac{w_{e,o_e}(t)}{\\eta^{out}_e} \
    + w_{i_e,e}(t) \\cdot \\eta^{in}_e  \\qquad \\forall e, \\forall t \\in [2, t_{max}]


    With :math:`e  \\in \mathcal{E}` and :math:`\mathcal{E}` beeing the
    set of unique ids for all entities grouped in the attribute `block.objs`.

    Additionally: :math:`\mathcal{E} \subset \mathcal{E}_{S}`.


    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class
    """
    if not block.objs or block.objs is None:
        raise ValueError('No objects defined. Please specify objects for' +
                         'which storage balance constraint should be set.')
    # constraint for storage energy balance
    cap_initial = {}
    cap_loss = {}
    eta_in = {}
    eta_out = {}

    for e in block.objs:
        cap_initial[e.uid] = e.cap_initial
        cap_loss[e.uid] = e.cap_loss
        # if cap_loss is no list or Series, alter to list
        if not isinstance(cap_loss[e.uid], (list, pdSeries)):
            cap_loss[e.uid] = [cap_loss[e.uid]]*len(model.timesteps)
        eta_in[e.uid] = e.eta_in
        eta_out[e.uid] = e.eta_out

    # set cap of last timesteps to fixed value of cap_initial
    t_last = len(model.timesteps)-1
    for e in block.uids:
        if cap_initial[e] is not None:
            block.cap[e, t_last] = cap_initial[e]
            block.cap[e, t_last].fix()

    def storage_balance_rule(block, e, t):
        # TODO: include time increment
        expr = 0
        if(t == 0):
            t_last = len(model.timesteps)-1
            expr += block.cap[e, t]
            expr += - block.cap[e, t_last] * (1 - cap_loss[e][t])
            expr += - model.w[model.I[e][0], e, t] * eta_in[e]
            expr += + model.w[e, model.O[e][0], t] / eta_out[e]
        else:
            expr += block.cap[e, t]
            expr += - block.cap[e, t-1] * (1 - cap_loss[e][t])
            expr += - model.w[model.I[e][0], e, t] * eta_in[e]
            expr += + model.w[e, model.O[e][0], t] / eta_out[e]
        return(expr, 0)
    block.balance = po.Constraint(block.indexset, rule=storage_balance_rule)


def add_storage_charge_discharge_limits(model, block):
    """ Constraints that limit the discharge and charge power by the c-rate

    Constraints are for investment models only.

    The mathematical formulation for the constraints is as follows:

    Discharge:

    .. math:: w_{e, o_e}(t) \\leq (\overline{L}_e + \overline{l}^{add}_e) \
        \\cdot C^{rate}_{o_e}
        \\qquad \\forall e, \\forall t

    Charge:

    .. math:: w_{i_e, e}(t) \\leq (\overline{L}_e + \overline{l}^{add}_e) \
        \\cdot C^{rate}_{i_e}
        \\qquad \\forall e, \\forall t


    """

    c_rate_out = {obj.uid: obj.c_rate_out for obj in block.objs}
    c_rate_in = {obj.uid: obj.c_rate_in for obj in block.objs}
    cap_max = {obj.uid: obj.cap_max for obj in block.objs}

    def storage_discharge_limit_rule(block, e, t):
        expr = 0
        expr += model.w[e, model.O[e][0], t]
        expr += -(cap_max[e] + block.add_cap[e]) \
            * c_rate_out[e]
        return(expr <= 0)
    block.discharge_limit_invest = po.Constraint(block.uids,
                                                 model.timesteps,
                                                 rule=
                                                 storage_discharge_limit_rule)

    def storage_charge_limit_rule(block, e, t):
        expr = 0
        expr += model.w[e, model.I[e][0], t]
        expr += -(cap_max[e] + block.add_cap[e]) \
            * c_rate_in[e]
        return(expr <= 0)
    block.charge_limit_invest = po.Constraint(block.uids,
                                              model.timesteps,
                                              rule=storage_charge_limit_rule)


def add_output_gradient_calc(model, block, grad_direc='both', idx=0):
    """ Add constraint to calculate the gradient between two timesteps
    (positive and negative)

    The mathematical formulation for constraints are as follows:

    Positive gradient:

    .. math::  w_{e,o_{e,n}}(t) - w_{e, o_{e,n}}(t-1) \\leq g^{pos}_{e_o}(t)\
    \\qquad \\forall e, \\forall t / t=1

    .. math:: g^{pos}_{e_o}(t) \\leq \overline{G}^{pos}_{e_o}, \\qquad \\forall e, \\forall t

    Negative gradient:

        .. math::  w_{e,o_{e,n}}(t-1) - w_{e,o_{e,n}}(t) \\leq g^{neg}_{e_o}(t)\
    \\qquad \\forall e, \\forall t / t=1

    .. math:: g^{neg}_{e_o}(t) \\leq \overline{G}^{neg}_{e_o}, \\qquad \\forall e, \\forall t

    With :math:`e  \\in \mathcal{E}` and :math:`\mathcal{E}` beeing the
    set of unique ids for all entities grouped in the attribute `block.objs`.

    Additionally: :math:`\mathcal{E} \subset \mathcal{E}_{C}`.

    :math:`n` indicates the n-th output of component :math:`e` (arg: idx)

    Parameters
    ----------
    model : OptimizationModel() instance
        An object to be solved containing all Variables, Constraints, Data.
    block : SimpleBlock()
         block to group all constraints and variables etc., block corresponds
         to one oemof base class

    grad_direc: string
        string defining the direction of the gradient constraint.
        ('positive', negative', 'both')

    """
    if not block.objs or block.objs is None:
        raise ValueError('No objects defined. Please specify objects for' +
                          'which output gradient constraints should be set.')

    def grad_pos_calc_rule(block, e, t):
        if t > 0:
            lhs = model.w[e, model.O[e][idx], t] - model.w[e,model.O[e][idx], t-1]
            rhs = block.grad_pos_var[e, t]
            return(lhs <= rhs)
        else:
            return(po.Constraint.Skip)

    def grad_neg_calc_rule(block, e, t):
        if t > 0:
            lhs = model.w[e, model.O[e][idx], t-1] - model.w[e,model.O[e][idx], t]
            rhs = block.grad_neg_var[e, t]
            return(lhs <= rhs)
        else:
            return(po.Constraint.Skip)

    def grad_pos_bound_rule(block, e, t):
        return((0, grad_pos[e]))

    def grad_neg_bound_rule(block, e, t):
        return((0, grad_neg[e]))

    # negative gradient
    if grad_direc == 'positive' or grad_direc == "both":
        # create variable
        grad_pos = {obj.uid: obj.grad_pos for obj in block.objs}
        block.grad_pos_var = po.Var(block.indexset, within=po.NonNegativeReals,
                                    bounds=grad_pos_bound_rule)
        # set constraint
        block.grad_pos_calc = po.Constraint(block.indexset,
                                            rule=grad_pos_calc_rule)

    # positive gradient
    if grad_direc == 'negative' or grad_direc == "both":
        # create variable
        grad_neg = {obj.uid: obj.grad_neg for obj in block.objs}
        block.grad_neg_var = po.Var(block.indexset, within=po.NonNegativeReals,
                                    bounds=grad_neg_bound_rule)
        # set constraint
        block.grad_neg_calc = po.Constraint(block.indexset,
                                            rule=grad_neg_calc_rule)


